# Encoding: utf-8
import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.constants as sc
from numpy.lib.scimath import sqrt
from scipy.linalg import expm as scipy_expm


def calc_pseudo_diel(rho, angle: float, output: str = "eps") -> pd.DataFrame:
    """Calculates the pseudo dielectric function of a measurement from rho.

    Args:
        rho (pandas.DataFrame):
            Measurement DataFrame containing rho as complex number as column and wavelength as index
        angle (float): Angle of measurement in degree
        output (str, optional): Output format for dielectric function.
            'n': refractive index,
            'eps': Dielectric function as two-column pandas.DataFrame,
            'epsi': Dielectric function as imaginary number.
            Defaults to 'eps'.

    Returns:
        pandas.DataFrame: Frame containing the pseudo dielectric function or refractive index.
    """
    theta = angle * np.pi / 180
    eps = np.sin(theta) ** 2 * (1 + np.tan(theta) ** 2 * ((1 - rho) / (1 + rho)) ** 2)

    if output == "n":
        n = sqrt(eps)
        return pd.DataFrame({"n": n.real, "k": n.imag}, index=eps.index)

    if output == "epsi":
        return eps

    return pd.concat(
        {"ϵ1": eps.apply(lambda x: x.real), "ϵ2": eps.apply(lambda x: x.imag)}, axis=1
    )


def calc_rho(psi_delta: pd.DataFrame) -> pd.DataFrame:
    """Calculate rho from a Psi-Delta DataFrame.
    The Psi-Delta DataFrame should be structured as follows:

        :index: Wavelength
        :column 'Ψ': Psi from measurement
        :column 'Δ': Delta from measurement

    This format is as returned from SpectraRay.read_psi_delta_file(...).

    Args:
        psi_delta (pandas.DataFrame): DataFrame containing Psi+Delta Measurement data

    Returns:
        pandas.DataFrame: Frame containing rho as an imaginary number.
    """
    return psi_delta.apply(
        lambda x: np.tan(np.deg2rad(x["Ψ"])) * np.exp(-1j * np.deg2rad(x["Δ"])), axis=1
    )


def convert_psi_delta_to_isotropic_mueller_matrix(psi_delta_df) -> pd.DataFrame:
    r"""Extract aois and wavelengths values from psi_delta pandas.DataFrame and convert it to a Muellermatrix;
    only works for isotropic media.

    Args:
            psi_delta_df (pd.DataFrame): dataFrame returned from data importers'

    Returns:
            pd.DataFrame: pyElli-compatible DataFrame of the Mueller matrix
    """
    aois = psi_delta_df.index.get_level_values("Angle of Incidence").unique().to_numpy()
    wavelengths = psi_delta_df.index.get_level_values("Wavelength").unique().to_numpy()

    # create empty Mueller matrix
    MM = pd.DataFrame(
        {},
        columns=[
            "M11",
            "M12",
            "M13",
            "M14",
            "M21",
            "M22",
            "M23",
            "M24",
            "M31",
            "M32",
            "M33",
            "M34",
            "M41",
            "M42",
            "M43",
            "M44",
        ],
        index=pd.MultiIndex.from_product(
            [aois, wavelengths], names=["Angle of Incidence", "Wavelength"]
        ),
        dtype=float,
    )

    # fill in the Mueller matrix coefficients for an isotropic material based equation in https://www.jawoollam.com/resources/ellipsometry-faq#toggle-id-15
    for aoi in aois:
        for λ in wavelengths:
            Δ = psi_delta_df.loc[aoi, λ]["Δ"]
            Ψ = psi_delta_df.loc[aoi, λ]["Ψ"]

            N = np.cos(2 * Ψ)
            C = np.sin(2 * Ψ) * np.cos(2 * Δ)
            S = np.sin(2 * Ψ) * np.cos(2 * Δ)

            MM.loc[(aoi, λ), "M11"] = 1
            MM.loc[(aoi, λ), "M12"] = -N
            MM.loc[(aoi, λ), "M13"] = 0
            MM.loc[(aoi, λ), "M14"] = 0

            MM.loc[(aoi, λ), "M21"] = -N
            MM.loc[(aoi, λ), "M22"] = 1
            MM.loc[(aoi, λ), "M23"] = 0
            MM.loc[(aoi, λ), "M24"] = 0

            MM.loc[(aoi, λ), "M31"] = 0
            MM.loc[(aoi, λ), "M32"] = 0
            MM.loc[(aoi, λ), "M33"] = C
            MM.loc[(aoi, λ), "M34"] = S

            MM.loc[(aoi, λ), "M41"] = 0
            MM.loc[(aoi, λ), "M42"] = 0
            MM.loc[(aoi, λ), "M43"] = -S
            MM.loc[(aoi, λ), "M44"] = C

    return MM


def get_qwp_thickness(material: "Material", lbda: float) -> float:
    """Return the thickness of a material in nm for a quarter wave plate at wavelength 'lbda'.

    Args:
        material (Material): Material object of the quarter wave plate
        lbda (float): Wavelength (in nm) at which the quarter wave plate is calculated

    Returns:
        float: Thickness (in nm) of quarter wave plate
    """
    nr = np.real(material.get_refractive_index(lbda)[0, 0, 0])
    return lbda / (4.0 * nr)


# base vectors
E_X = np.array([1, 0, 0]).reshape((3,))
E_Y = np.array([0, 1, 0]).reshape((3,))
E_Z = np.array([0, 0, 1]).reshape((3,))


def conversion_wavelength_energy(value: npt.ArrayLike) -> npt.ArrayLike:
    r"""Converts wavelength values to energy values and vice versa.

    .. math::
        value_{\text{target}} = c \cdot \hbar / \boldsymbol{value}

    Args:
        value (npt.ArrayLike): Single value or array of wavelengths in nm or energy in eV.

    Returns:
        npt.ArrayLike: Energy in eV or wavelength in nm.
    """
    return sc.speed_of_light * sc.value("Planck constant in eV/Hz") / (value * 1e-9)


def conversion_frequency2energy(f: npt.ArrayLike) -> npt.ArrayLike:
    r"""Converts frequency values to energy values.

    .. math::
        E = \boldsymbol{f} \cdot \hbar

    Args:
        f (npt.ArrayLike): Single value or array of frequencies in Hz.

    Returns:
        npt.ArrayLike: Energy in eV.
    """
    return f * sc.value("Planck constant in eV/Hz")


def conversion_energy2frequency(E: npt.ArrayLike) -> npt.ArrayLike:
    r"""Converts energy values to frequency values.

    .. math::
        f = \boldsymbol{E} / \hbar

    Args:
        E (npt.ArrayLike): Single value or array of energies in eV.

    Returns:
        npt.ArrayLike: Frequency in Hz.
    """
    return E / sc.value("Planck constant in eV/Hz")


def conversion_wavelength_frequency(value: npt.ArrayLike) -> npt.ArrayLike:
    r"""Converts wavelength values to frequency values and vice versa.

    .. math::
        value_{\text{target}} = c / \boldsymbol{value}

    Args:
        value (npt.ArrayLike): Single value or array of wavelengths in nm or frequencies in Hz.

    Returns:
        npt.ArrayLike: Frequencies in Hz or wavelengths in nm.
    """
    return sc.speed_of_light / (value * 1e-9)


def conversion_wavelength_wavenumber(value: npt.ArrayLike) -> npt.ArrayLike:
    r"""Converts wavelength values to wavenumber values and vice versa.

    .. math::
        value_{\text{target}} = 1 / \boldsymbol{value}

    Args:
        value (npt.ArrayLike): Single value or array of wavelengths in nm
            or wavenumbers in :math:`\text{cm}^{-1}`.

    Returns:
        npt.ArrayLike: Wavenumbers in :math:`\text{cm}^{-1}` or wavelengths in nm.
    """
    return 1e7 / value


#########################################################
# Rotations


def rotation_euler(p: float, n: float, r: float) -> npt.NDArray:
    """Returns rotation matrix defined by Euler angles p, n, r.

    Successive rotations : z,x',z'
    Note : The inverse rotation is -r, -n, -p

    Args:
        p (float): precession angle, 1st rotation, around z (0..360°).
        n (float): nutation angle, 2nd rotation, around x' (0..180°).
        r (float): 3rd rotation, around z' (0..360°).

    Returns:
        npt.NDArray: rotation matrix :math:`M_R`
    """
    p = np.deg2rad(p)
    n = np.deg2rad(n)
    r = np.deg2rad(r)

    c1 = np.cos(p)
    s1 = np.sin(p)
    c2 = np.cos(n)
    s2 = np.sin(n)
    c3 = np.cos(r)
    s3 = np.sin(r)

    return np.array(
        [
            [c1 * c3 - s1 * c2 * s3, -c1 * s3 - s1 * c2 * c3, s1 * s2],
            [s1 * c3 + c1 * c2 * s3, -s1 * s3 + c1 * c2 * c3, -c1 * s2],
            [s2 * s3, s2 * c3, c2],
        ]
    )


def rotation_v(v: npt.ArrayLike) -> npt.NDArray:
    r"""Returns rotation matrix defined by a rotation vector v.

    The calculation is made with the matrix exponential
    :math:`M_R = \exp(W)`, with :math:`W_{ij} = - ε_{ijk} V_{k}`,
    where :math:`ε_{ijk}` is the Levi-Civita antisymmetric tensor.
    If V is separated in a unit vector v and a magnitude θ, V = θ·v, with
    θ = ∥V∥, the calculation of the matrix exponential is avoided, and only
    sin(θ) and cos(θ) are needed instead.

    Note : The inverse rotation is -v

    Args:
        v (npt.ArrayLike): rotation vector (list or array)

    Returns:
        npt.NDArray: rotation matrix :math:`M_R`
    """
    # fmt: off
    m_w = np.array([[0,     -v[2], v[1]],
                  [v[2],  0,     -v[0]],
                  [-v[1], v[0],  0]])
    # fmt: on

    return scipy_expm(m_w)


def rotation_v_theta(v: npt.ArrayLike, theta: float) -> npt.NDArray:
    """Returns rotation matrix defined by a unit rotation vector and an angle.

    Notes : The inverse rotation is (v,-theta)

    Args:
        v (npt.ArrayLike): unit vector orienting the rotation (list or array)
        theta (float): rotation angle around v in degrees

    Returns:
        npt.NDArray: rotation matrix :math:`M_R`
    """
    # fmt: off
    m_w = np.array([[0,     -v[2], v[1]],
                  [v[2],  0,     -v[0]],
                  [-v[1], v[0],  0]])
    # fmt: on

    return (
        np.identity(3)
        + m_w * np.sin(np.deg2rad(theta))
        + np.linalg.matrix_power(m_w, 2) * (1 - np.cos(np.deg2rad(theta)))
    )



# Encoding: utf-8
"""Dispersion law for an UV and IR pole."""

import numpy.typing as npt

from ..utils import conversion_wavelength_energy
from .base_dispersion import Dispersion


class Poles(Dispersion):
    r"""Dispersion law for an UV and IR pole,
    i.e. Lorentz oscillators outside the fitting spectral range and zero broadening.

    Single parameters:
        :A_ir: IR Pole amplitude. Defaults to 1. Unit in eV\ :sup:`2`.
        :A_uv: UV Pole amplitude. Defaults to 1. Unit in eV\ :sup:`2`.
        :E_uv: UV Pole energy. Defaults to 6. Unit in eV.

    Repeated parameters:
        --

    Output:
        .. math::
                \varepsilon(E) = \boldsymbol{A\_ir} / E^2
                + \boldsymbol{A\_uv} / (\boldsymbol{E\_uv}^2 - E^2)
    """

    single_params_template = {"A_ir": 1, "A_uv": 1, "E_uv": 6}
    rep_params_template = {}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        energy = conversion_wavelength_energy(lbda)
        return self.single_params.get("A_ir") / energy**2 + self.single_params.get(
            "A_uv"
        ) / (self.single_params.get("E_uv") ** 2 - energy**2)



"""
Multilayer fit
==============

Fits a multilayer model to an ALD grown TiO2 sample on SiO2 / Si.
"""

# %%
import elli
from elli.fitting import ParamsHist, fit

# sphinx_gallery_thumbnail_path = '_static/multilayer.png'

# %%
# Load data
# ---------
#
# Load data collected with Sentech Ellipsometer and cut the spectral range (to use Si Aspnes file)
#
# The sample is an ALD grown TiO2 sample (with 400 cycles)
# on commercially available SiO2 / Si substrate.
tss = elli.read_spectraray_psi_delta("TiO2_400cycles.txt").loc[70.06][400:800]

# %%
# Set start parameters
# --------------------
# Here we set the start parameters for the TiO2 and SiO2 layer.
# We set the SiO2 layer parameters to a fixed value from another
# fit of the substrate. See the :ref:`Basic usage` example for details
# on how to perform such a fit.
# In general it is a good idea to fit your data layer-wise if possible
# to yield a better fit quality.
params = ParamsHist()
params.add("SiO2_n0", value=1.452, min=-100, max=100, vary=False)
params.add("SiO2_n1", value=36.0, min=-40000, max=40000, vary=False)
params.add("SiO2_n2", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_k0", value=0, min=-100, max=100, vary=False)
params.add("SiO2_k1", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_k2", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_d", value=276.36, min=0, max=40000, vary=False)

params.add("TiO2_n0", value=2.236, min=-100, max=100, vary=True)
params.add("TiO2_n1", value=451, min=-40000, max=40000, vary=True)
params.add("TiO2_n2", value=251, min=-40000, max=40000, vary=True)
params.add("TiO2_k0", value=0, min=-100, max=100, vary=False)
params.add("TiO2_k1", value=0, min=-40000, max=40000, vary=False)
params.add("TiO2_k2", value=0, min=-40000, max=40000, vary=False)

params.add("TiO2_d", value=20, min=0, max=40000, vary=True)

# %%
# Load silicon dispersion from the refractiveindexinfo database
# -------------------------------------------------------------
# You can load any material from the index
# `refractiveindex.info <https://refractiveindex.info>`__, which is
# embedded into the software (so you may use it offline, too). Here, we
# are interested in the literature values for the silicon substrate.
# First we need to load the database with ``rii_db = elli.db.RII()`` and
# then we can query it with ``rii_db.get_mat("Si", "Aspnes")`` to load
# this
# `entry <https://refractiveindex.info/?shelf=main&book=Si&page=Aspnes>`__.
rii_db = elli.db.RII()
Si = rii_db.get_mat("Si", "Aspnes")


# %%
# Building the model
# ------------------
# Here the model is build and the experimental structure is returned.
# For details on this process please refer to the :ref:`Basic usage` example.
# When executed in an jupyter notebook this displays an interactive graph
# with which you can select the start parameters before fitting the data.
@fit(tss, params)
def model(lbda, params):
    SiO2 = elli.Cauchy(
        params["SiO2_n0"],
        params["SiO2_n1"],
        params["SiO2_n2"],
        params["SiO2_k0"],
        params["SiO2_k1"],
        params["SiO2_k2"],
    ).get_mat()
    TiO2 = elli.Cauchy(
        params["TiO2_n0"],
        params["TiO2_n1"],
        params["TiO2_n2"],
        params["TiO2_k0"],
        params["TiO2_k1"],
        params["TiO2_k2"],
    ).get_mat()

    Layer = [elli.Layer(TiO2, params["TiO2_d"]), elli.Layer(SiO2, params["SiO2_d"])]

    return elli.Structure(elli.AIR, Layer, Si).evaluate(lbda, 70, solver=elli.Solver2x2)
    # Alternative: Use 4x4 Solver with scipy propagator
    # return elli.Structure(elli.AIR, Layer, Si).evaluate(lbda, 70, solver=elli.Solver4x4, propagator=elli.PropagatorExpm())


# %%
# Plot & Fit model
# ----------------
# We plot the model to see the deviation with the initial parameters.
model.plot()


# %%
# Now lets perform the fit and plot the comparison of
# calculation and experimental data afterwards.
fit_stats = model.fit()
model.plot()

# %%
# We can also have a look at the fit statistics.
fit_stats

# %%
# References
# ----------
# `Here <https://github.com/PyEllips/pyElli/tree/master/examples/TiO2%20Fit>`_
# you can find the latest jupyter notebook and data files of this example.



"""Read and write NeXus files in the ellipsometry
`standard <https://manual.nexusformat.org/classes/contributed_definitions/NXellipsometry.html#nxellipsometry>`_.
NeXus is a format originating from large beam line facitilies and is adapted for
other smaller labratory experiments to supply an agreed standard for data sharing.
For now only reading is supported but in the future there will also be a writer
to store the whole optical model and fit inside the NeXus file."""

from dataclasses import dataclass
from typing import Callable, Optional

import h5py
import numpy as np
import pandas as pd

if np.lib.NumpyVersion(np.__version__) < "2.0.0":
    from numpy.lib.index_tricks import IndexExpression
else:
    from numpy.lib._index_tricks_impl import IndexExpression

from elli.dispersions.formula import Formula, FormulaIndex
from elli.dispersions.table_epsilon import TableEpsilon
from elli.dispersions.table_index import Table
from elli.materials import BiaxialMaterial, IsotropicMaterial, UniaxialMaterial

from ..utils import calc_rho, conversion_wavelength_energy


@dataclass
class NexusGroupNames:
    """Contains nexus group names to read from the nexus file"""

    entry: str = "entry"
    sample: str = "sample"
    instrument: str = "instrument"

    @property
    def full_instrument_path(self):
        """Get the full instrument path with prepended entry name, e.g. entry/instrument"""
        return f"{self.entry}/{self.instrument}"

    @property
    def full_sample_path(self):
        """Get the full sample path with prepended entry name, e.g. entry/sample"""
        return f"{self.entry}/{self.sample}"


def read_nexus_psi_delta(
    nxs_filename: str, group_names: Optional[NexusGroupNames] = None
) -> pd.DataFrame:
    """Read a NeXus file containing Psi and Delta data.

    Args:
        nxs_filename (str): The filename of the NeXus file.

    Raises:
        ValueError: Is raised when the data is not stored as psi / delta value.

    Returns:
        pd.DataFrame: Psi/Delta DataFrame.
        The index is a multiindex consisting of the angle of incidents as first column
        and the wavelength as second column.
    """

    def read_data(
        wavelength_path: str,
        aois_path: str,
        data_path: str,
        indexing: Callable[[int, int], IndexExpression],
    ):
        aois = np.array(h5file[aois_path])
        wavelength = np.array(h5file[wavelength_path]) / 10
        psi_delta_df = pd.DataFrame(
            {},
            columns=["Ψ", "Δ"],
            index=pd.MultiIndex.from_product(
                [aois, wavelength], names=["Angle of Incidence", "Wavelength"]
            ),
            dtype=float,
        )

        data = np.array(h5file[data_path])

        for i, aoi in enumerate(aois):
            psi_delta_df.loc[aoi, "Ψ"] = data[indexing(i, 0)]
            psi_delta_df.loc[aoi, "Δ"] = data[indexing(i, 1)]

        return psi_delta_df

    def read_legacy() -> pd.DataFrame:
        return read_data(
            wavelength_path=f"{group_names.full_instrument_path}/spectrometer/wavelength",
            aois_path=f"{group_names.full_instrument_path}/angle_of_incidence",
            data_path=f"{group_names.full_sample_path}/measured_data",
            indexing=lambda aoi_idx, observable_idx: np.s_[
                0, 0, aoi_idx, observable_idx, :
            ],
        )

    def read_nx_opt_def() -> pd.DataFrame:
        return read_data(
            wavelength_path=f"{group_names.entry}/data_collection/wavelength_spectrum",
            aois_path=f"{group_names.full_instrument_path}/angle_of_incidence",
            data_path=f"{group_names.entry}/data_collection/measured_data",
            indexing=lambda aoi_idx, observable_idx: np.s_[aoi_idx, observable_idx, :],
        )

    if group_names is not None and not isinstance(group_names, NexusGroupNames):
        raise ValueError(
            f"Invalid type for for group_names: {type(group_names)}. "
            "Should be an instance of NexusGroupNames."
        )

    if group_names is None:
        group_names = NexusGroupNames()

    h5file = h5py.File(nxs_filename, "r")
    if f"{group_names.entry}/sample/data_type" in h5file:
        data_type = h5file[f"{group_names.entry}/sample/data_type"][()].decode("utf-8")
    elif f"{group_names.entry}/data_collection/data_type" in h5file:
        data_type = h5file[f"{group_names.entry}/data_collection/data_type"][()].decode(
            "utf-8"
        )
    else:
        raise ValueError(
            "Could not resolve a proper definition "
            f"from the provided nexus file: {nxs_filename}"
        )

    # Currently, the appdef version can only be determined
    # reliably by the case in the data_type field.
    def_mapping = {
        "psi/delta": read_legacy,
        "Psi/Delta": read_nx_opt_def,
    }

    if data_type not in def_mapping:
        raise NotImplementedError(
            f"Unsupported data type: {data_type}. "
            "Only 'psi/delta' values are supported yet."
        )

    return def_mapping.get(data_type)()


def read_nexus_rho(nxs_filename: str) -> pd.DataFrame:
    """Reads rho value from NeXus datafile.
    Currently, this works only with psi / delta representation in the NeXus file.

    Raises:
        ValueError: Is raised when the data is not stored as psi / delta value.

    Returns:
        pd.DataFrame: DataFrame containing the measured data as imaginary rho value.
        The index is a multiindex consisting of the angle of incidents as first column
        and the wavelength as second column.
    """
    return calc_rho(read_nexus_psi_delta(nxs_filename))


def read_nexus_materials(filename: str):
    """Read the optical materials from a nexus file.

    Args:
        filename (str): The nexus filename.
    """

    def get_dispersion_function(dataset: h5py.Dataset):
        single_params = {}
        rep_params = {}

        def get_params(_: str, dataset: h5py.Dataset):
            if dataset.attrs.get("NX_class", "") in ["NXdispersion_single_parameter"]:
                param_name = dataset["name"][()].decode("utf-8")
                single_params[param_name] = dataset["value"][()]

            if dataset.attrs.get("NX_class", "") in ["NXdispersion_repeated_parameter"]:
                param_name = dataset["name"][()].decode("utf-8")
                rep_params[param_name] = dataset["values"][()]

        identifier = None
        if "wavelength_identifier" in dataset:
            identifier = dataset["wavelength_identifier"][()].decode("utf-8")
            unit = (
                f"{dataset['wavelength_unit'][()]} "
                f"{dataset['wavelength_unit'].attrs.get('units', '')}"
            )

        if "energy_identifier" in dataset:
            raise NotImplementedError(
                "Using dispersions with an energy axis is not yet supported."
            )

        representation = dataset["representation"][()].decode("utf-8")
        formula = dataset["formula"][()].decode("utf-8")
        dataset.visititems(get_params)

        if representation == "eps":
            return Formula(formula, identifier, single_params, rep_params, unit)

        if representation == "n":
            return FormulaIndex(formula, identifier, single_params, rep_params, unit)

        raise ValueError(f"Unsupported representation {representation}")

    def get_dispersion_table(dataset: h5py.Dataset):
        wavelength = None
        if "wavelength" in dataset:
            wavelength = dataset["wavelength"]

        if "energy" in dataset:
            wavelength = conversion_wavelength_energy(dataset["energy"])

        if wavelength is None:
            raise ValueError("No wavelength array found in dataset.")

        if "refractive_index" in dataset:
            return Table(lbda=wavelength, n=dataset["refractive_index"][()])

        if "dielectric_function" in dataset:
            return TableEpsilon(
                lbda=wavelength, epsilon=dataset["dielectric_function"][()]
            )

        raise ValueError(
            "Invalid dispersion table, neither `refractive_index` "
            "nor `dielectric_function` found."
        )

    def get_dispersion(dataset: h5py.Dataset):
        dispersion = None
        for data in dataset.values():
            if data.attrs.get("NX_class") in ["NXdispersion_function"]:
                dispersion = (
                    get_dispersion_function(data)
                    if dispersion is None
                    else dispersion + get_dispersion_function(data)
                )

            if data.attrs.get("NX_class") in ["NXdispersion_table"]:
                dispersion = (
                    get_dispersion_table(data)
                    if dispersion is None
                    else dispersion + get_dispersion_table(data)
                )

        return dispersion

    def get_material(dataset: h5py.Dataset):
        dispersions = {}
        for dispersion in [f"dispersion_{axis}" for axis in ["x", "y", "z"]]:
            if dispersion not in dataset:
                if dispersion == "dispersion_x":
                    raise ValueError("Invalid dispersion file. `dispersion_x` missing")
                continue
            dispersions[dispersion] = get_dispersion(dataset[dispersion])

        if len(dispersions) == 1:
            return IsotropicMaterial(dispersions["dispersion_x"])
        if len(dispersions) == 2:
            return UniaxialMaterial(
                dispersions["dispersion_x"], dispersions["dispersion_z"]
            )
        if len(dispersions) == 3:
            return BiaxialMaterial(**dispersions)

        raise ValueError("Could not create material from nexus file.")

    entries = {}
    with h5py.File(filename) as h5file:
        for entry, data in h5file.items():
            entries[entry] = get_material(data)

    return entries



"""ParmsHist provides a wrapper around lmfit.Parameters
to keep track of the changes made to the parameters."""

# Encoding: utf-8
import copy
from typing import List

try:
    from lmfit import Parameters
except ImportError as e:
    raise ImportError(
        "This module requires lmfit to work properly.\n"
        "Try installing this package with the additional fitting requirement, "
        "i.e. pip install pyElli[fitting]"
    ) from e


class ParamsHist(Parameters):
    """A wrapper around lmfit.Parameters to keep track of the changes made to the parameters."""

    def __init__(self) -> None:
        super().__init__()
        self._history = []
        self._max_length = 50

    @property
    def history(self) -> List[Parameters]:
        """Gets the entire history

        Returns:
            List[Parameters]: The history
        """
        return self._history

    def clear_history(self) -> None:
        """Clears the parameters history"""
        self._history = []

    @property
    def history_len(self) -> int:
        """The current length of the history

        Returns:
            int: The length of the history
        """
        return len(self._history)

    @property
    def max_history_len(self) -> int:
        """The maximum length of the history

        Returns:
            int: The maximum length of the history
        """
        return self._max_length

    @max_history_len.setter
    def max_history_len(self, history_len: int) -> None:
        """Sets the maximum history length. If the current
        history length is greater than the new history length the history
        gets truncated.

        Args:
            history_len (int): The new history length

        Raises:
            ValueError: If history_len is not an int or < 1.
        """
        if not isinstance(history_len, int):
            raise ValueError("History length has to be an integer")

        if history_len < 1:
            raise ValueError("History length must be greater than 0")

        self._history = self._history[-history_len:]
        self._max_length = history_len

    def revert(self, hist_pos: int) -> None:
        """Reverts to an older history version and keeps the entire history.

        Args:
            hist_pos (int): The history position to revert to.
        """
        if len(self._history) > (hist_pos % len(self._history)):
            self.update(self._history[hist_pos])

    def pop(self):
        """Gets to the previous history version and deletes the current element."""
        if len(self._history) > 0:
            curr_params = self.copy()
            self.update(self._history[-1])
            self._history = self._history[:-1]

            return curr_params
        return None

    def update_value(self, key: str, value: float) -> None:
        """Updates a parameter and keeps track of the change in history

        Args:
            key (str): The key to be updated
            value (float): The value the key should be updated to
        """
        self.commit()
        super().__getitem__(key).value = value

    def update_params(self, parameters) -> None:
        """Updates the current parameters from a lmfit parameters object.

        Args:
            parameters (lmfit.Parameters):
                The lmfit parameters object to update the values from.
        """
        self.commit()
        self.update(parameters)

    def tracked_add(self, *args, **kwargs) -> None:
        """Adds a parameter and keeps track of the change in history"""
        self.commit()
        super().add(*args, **kwargs)

    def commit(self) -> None:
        """Saves the current parameter set to history."""
        if len(self._history) >= self._max_length:
            self._history = self._history[1:]
        clone = copy.deepcopy(self)
        self.history.append(clone)



"""
Cholesteric Liquid Crystal
==========================
"""

# %%
# Example of a cholesteric liquid crystal
# ---------------------------------------
# Authors: O. Castany, C. Molinaro, M. Müller

# %%
import elli
import elli.plot as elliplot
import matplotlib.pyplot as plt
import numpy as np
from scipy.constants import c, pi

# %%
# Setup materials and structure
# -----------------------------
glass = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(1.55))
front = back = glass

# Liquid crystal oriented along the x direction
(no, ne) = (1.5, 1.7)
Dn = ne - no
n_med = (ne + no) / 2
LC = elli.UniaxialMaterial(
    elli.ConstantRefractiveIndex(no), elli.ConstantRefractiveIndex(ne)
)  # ne is along z
R = elli.rotation_v_theta(elli.E_Y, 90)  # rotation of pi/2 along y
LC.set_rotation(R)  # apply rotation from z to x

# Cholesteric pitch (nm):
p = 650

# One half turn of a right-handed helix:
TN = elli.TwistedLayer(LC, p / 2, angle=180, div=35)

# Repetition the helix layer
N = 15  # number half pitch repetitions
h = N * p / 2
L = elli.RepeatedLayers([TN], N)
s = elli.Structure(front, [L], back)

# Calculation parameters
lbda_min, lbda_max = 800, 1200  # (nm)
lbda_B = p * n_med
lbda_list = np.linspace(lbda_min, lbda_max, 100)

# %%
# Analytical calculation for the maximal reflection
# -------------------------------------------------
R_th = np.tanh(Dn / n_med * pi * h / p) ** 2
lbda_B1, lbda_B2 = p * no, p * ne

# %%
# Calculation with pyElli
# -----------------------
data = s.evaluate(lbda_list, 0)

T_pp = data.T_pp
T_ps = data.T_ps
T_ss = data.T_ss
T_sp = data.T_sp

# Transmission coefficients for incident unpolarized light:
T_pn = 0.5 * (T_pp + T_ps)
T_sn = 0.5 * (T_sp + T_ss)
T_nn = T_sn + T_pn

# Transmission coefficients for 's' and 'p' polarized light, with
# unpolarized measurement.
T_ns = T_ps + T_ss
T_np = T_pp + T_sp

# Right-circular wave is reflected in the stop-band.
# R_LR, T_LR close to zero.
R_RR = data.Rc_RR
R_LR = data.Rc_LR
T_RR = data.Tc_RR
T_LR = data.Tc_LR

# Left-circular wave is transmitted in the full spectrum.
# T_RL, R_RL, R_LL close to zero, T_LL close to 1.
T_LL = data.Tc_LL
R_LL = data.Rc_LL

# %%
# Plotting
# --------
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)

# Draw rectangle for λ ∈ [p·no, p·ne], and T ∈ [0, R_th]
rectangle = plt.Rectangle((lbda_B1, 0), lbda_B2 - lbda_B1, R_th, color="cyan")
ax.add_patch(rectangle)

ax.plot(lbda_list, R_RR, "--", label="R_RR")
ax.plot(lbda_list, T_RR, label="T_RR")
ax.plot(lbda_list, T_nn, label="T_nn")
ax.plot(lbda_list, T_ns, label="T_ns")
ax.plot(lbda_list, T_np, label="T_np")

ax.legend(loc="center right", bbox_to_anchor=(1.00, 0.50))

ax.set_title(
    "Right-handed Cholesteric Liquid Crystal, aligned along \n"
    + "the $x$ direction, with {:.1f} helix pitches.".format(N / 2.0)
)
ax.set_xlabel(r"Wavelength $\lambda_0$ (nm)")
ax.set_ylabel(r"Power transmission $T$ and reflexion $R$")
plt.show()

elliplot.draw_structure(s)



# Encoding: utf-8
"""Abstract base class and utility classes for pyElli dispersion"""

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import List, Optional, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from numpy.lib.scimath import sqrt

from .. import dispersions


class InvalidParameters(Exception):
    """Exception for invalid dispersion parameters."""


class BaseDispersion(ABC):
    """BaseDispersion (abstract class).

    Functions provided for derived classes:
    * dielectric_function(lbda) : returns dielectric constant for wavelength 'lbda'
    """

    default_lbda_range = np.linspace(200, 1000, 801)

    @property
    @abstractmethod
    def single_params_template(self) -> dict:
        """Specifies the single parameters of the model and its default values."""

    @property
    @abstractmethod
    def rep_params_template(self) -> dict:
        """Specifies the repeated parameters of the model and its default values."""

    @staticmethod
    def _guard_invalid_params(params1, params2):
        missing_params = np.array(params1)[np.where(~np.isin(params1, params2))]

        if len(missing_params) > 0:
            missing_param_strings = ", ".join(f"{p}" for p in missing_params)
            raise InvalidParameters(f"Invalid parameter(s): {missing_param_strings}")

    @staticmethod
    def _fill_params_dict(template: dict, *args, **kwargs) -> dict:
        BaseDispersion._guard_invalid_params(list(kwargs.keys()), list(template.keys()))

        if (len(kwargs) + len(args)) > len(template):
            raise InvalidParameters("Too many parameters")

        params = template.copy()
        pos_arguments = set()

        for i, val in enumerate(args):
            key = list(template.keys())[i]
            params[key] = val
            pos_arguments.add(key)

        for key, value in kwargs.items():
            if key in pos_arguments:
                raise InvalidParameters(
                    f"Parameter {key} already set by positional argument"
                )
            params[key] = value

        return params

    def __init__(self, *args, **kwargs):
        super()
        self.rep_params = []

        self.single_params = self._fill_params_dict(
            self.single_params_template, *args, **kwargs
        )

        for param in self.single_params:
            if self.single_params[param] is None:
                raise InvalidParameters(f"Please specify parameter {param}")

    @abstractmethod
    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        """Calculates the dielectric function in a given wavelength window.

        Args:
            lbda (npt.ArrayLike): The wavelength window with unit nm.

        Returns:
            npt.NDArray: The dielectric function for each wavelength point.
        """

    def get_mat(self):
        """Returns this dispersion as an isotropic material"""
        from ..materials import IsotropicMaterial

        return IsotropicMaterial(self)

    def add(self, *args, **kwargs) -> "BaseDispersion":
        """Adds a set of parameters to the dispersion.

        Returns:
            Dispersion: The current object with the additional parameters added.
        """
        rep_param_set = self._fill_params_dict(
            self.rep_params_template, *args, **kwargs
        )
        self.rep_params.append(rep_param_set)

        return self

    def get_dielectric(self, lbda: Optional[npt.ArrayLike] = None) -> npt.NDArray:
        """Returns the dielectric constant for wavelength 'lbda' default unit (nm)
        in the convention ε1 + iε2."""
        lbda = self.default_lbda_range if lbda is None else lbda
        return np.asarray(self.dielectric_function(lbda), dtype=np.complex128)

    def get_refractive_index(self, lbda: Optional[npt.ArrayLike] = None) -> npt.NDArray:
        """Returns the refractive index for wavelength 'lbda' default unit (nm)
        in the convention n + ik."""
        lbda = self.default_lbda_range if lbda is None else lbda

        if isinstance(self, IndexDispersion):
            return self.refractive_index(lbda)
        return sqrt(self.dielectric_function(lbda))

    def get_dielectric_df(
        self, lbda: Optional[npt.ArrayLike] = None, conjugate=False
    ) -> pd.DataFrame:
        """Returns the dielectric function as a pandas dataframe

        Args:
            lbda (npt.ArrayLike, optional): The wavelength range to use.
                If this parameter is not given a default spectral range
                from 200 to 1000 nm with 801 points is used.
                Defaults to None.
            conjugate (bool, optional): Per default the convention ε1 + iε2 is returned.
                If this flag is set to True, the ε1 + iε2 convention is returned.
                Defaults to False.

        Returns:
            pd.DataFrame:
                A pandas dataframe containing the wavelength as index
                and two rows containing ε1 and ε2.
        """
        lbda = self.default_lbda_range if lbda is None else lbda
        eps = self.get_dielectric(lbda)

        return pd.DataFrame(
            {"ϵ1": eps.real, "ϵ2": -eps.imag if conjugate else eps.imag},
            index=pd.Index(lbda, name="Wavelength"),
        )

    def get_refractive_index_df(
        self, lbda: Optional[npt.ArrayLike] = None, conjugate=False
    ) -> pd.DataFrame:
        """Returns the refractive index as a pandas dataframe

        Args:
            lbda (npt.ArrayLike, optional): The wavelength range to use.
                If this parameter is not given a default spectral range
                from 200 to 1000 nm with 801 points is used.
                Defaults to None.
            conjugate (bool, optional): Per default the convention n + ik is returned.
                If this flag is set to True, the n + ik convention is returned.
                Defaults to False.

        Returns:
            pd.DataFrame:
                A pandas dataframe containing the wavelength as index
                and two rows containing n and k.
        """
        lbda = self.default_lbda_range if lbda is None else lbda
        nk = self.get_refractive_index(lbda)

        return pd.DataFrame(
            {"n": nk.real, "k": -nk.imag if conjugate else nk.imag},
            index=pd.Index(lbda, name="Wavelength"),
        )

    def __repr__(self):
        def _dict_to_str(dic):
            return ", ".join(f"{item[0]} = {item[1]}" for item in dic.items())

        return (
            type(self).__name__
            + "\n"
            + "=" * len(type(self).__name__)
            + "\n"
            + _dict_to_str(self.single_params)
            + (
                "\n\nOscillators\n"
                + "===========\n"
                + "\n".join(_dict_to_str(p) for p in self.rep_params)
                if len(self.rep_params) > 0
                else ""
            )
        )


class Dispersion(BaseDispersion):
    """A dispersion based on a dielectric function formulation."""

    def _check_valid_operand(self, other: Union[int, float, "Dispersion"]):
        if not isinstance(other, (int, float, Dispersion, dispersions.TableEpsilon)):
            raise TypeError(
                f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'"
            )

    def __radd__(self, other: Union[int, float, "Dispersion"]) -> "DispersionSum":
        """Add up the dielectric function of multiple models"""
        return self.__add__(other)

    def __add__(self, other: Union[int, float, "Dispersion"]) -> "DispersionSum":
        """Add up the dielectric function of multiple models"""
        if isinstance(other, IndexDispersion):
            raise TypeError(
                "Cannot add refractive index and dielectric function based dispersions."
            )
        self._check_valid_operand(other)

        if isinstance(other, dispersions.TableEpsilon):
            return other.__add__(self)

        if isinstance(other, DispersionSum):
            other.dispersions.append(self)
            return other

        if isinstance(other, (int, float)):
            return DispersionSum(self, dispersions.EpsilonInf(other))

        return DispersionSum(self, other)

    def as_index(self):
        """
        Returns this class as IndexDispersion.
        This method may be used to add dielectric and index based dispersions.
        Please ensure that you know what you are doing as building dielectric
        and index based dispersions is normally mathematically wrong.
        """
        index_class = deepcopy(self)
        # pylint: disable=attribute-defined-outside-init
        index_class.refractive_index = lambda lbda: sqrt(
            index_class.dielectric_function(lbda)
        )
        index_class.__class__ = IndexDispersion  # pylint: disable=invalid-name
        index_class.dielectric_function = self.dielectric_function

        return index_class


class IndexDispersion(BaseDispersion):
    """A dispersion based on a refractive index formulation."""

    @abstractmethod
    def refractive_index(self, lbda: npt.ArrayLike) -> npt.NDArray:
        """Calculates the refractive index in a given wavelength window.

        Args:
            lbda (npt.ArrayLike): The wavelength window with unit nm.

        Returns:
            npt.NDArray: The refractive index for each wavelength point.
        """

    def _check_valid_operand(self, other: Union[int, float, "IndexDispersion"]):
        if not isinstance(other, (int, float, IndexDispersion, dispersions.Table)):
            raise TypeError(
                f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'"
            )

    def __radd__(
        self, other: Union[int, float, "IndexDispersion"]
    ) -> "IndexDispersionSum":
        """Add up the dielectric function of multiple models"""
        return self.__add__(other)

    def __add__(
        self, other: Union[int, float, "IndexDispersion"]
    ) -> "IndexDispersionSum":
        if isinstance(other, Dispersion):
            raise TypeError(
                "Cannot add refractive index and dielectric function based dispersions."
            )

        self._check_valid_operand(other)

        if isinstance(other, dispersions.Table):
            return other.__add__(self)

        if isinstance(other, IndexDispersionSum):
            other.index_dispersions.append(self)
            return other

        if isinstance(other, (int, float)):
            return IndexDispersionSum(self, dispersions.ConstantRefractiveIndex(other))

        return IndexDispersionSum(self, other)

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return self.refractive_index(lbda) ** 2

    def as_dielectric(self):
        """
        Returns this class as Dispersion.
        This method may be used to add dielectric and index based dispersions.
        Please ensure that you know what you are doing as building dielectric
        and index based dispersions is normally mathematically wrong.
        """
        diel_disp = deepcopy(self)
        diel_disp.__class__ = Dispersion  # pylint: disable=invalid-name
        diel_disp.dielectric_function = self.dielectric_function
        return diel_disp


class DispersionFactory:
    """A factory class for dispersion objects"""

    @staticmethod
    def get_dispersion(identifier: str, *args, **kwargs) -> Dispersion:
        """Creates a Dispersion object identified by its string name and initializes it
        with the given parameters.

        Args:
            identifier (str): Identifier of the Dispersion object, e.g. Cauchy.

        Returns:
            DispersionLaw: The Dispersion object initialized with the given parameters.
        """
        bad_identifier = ["Dispersion"]

        if hasattr(dispersions, identifier) and identifier not in bad_identifier:
            return getattr(dispersions, identifier)(*args, **kwargs)

        raise ValueError(f"No such dispersion: {identifier}")


class DispersionSum(Dispersion):
    """Represents a sum of two dispersions"""

    single_params_template: dict = {}
    rep_params_template: dict = {}
    dispersions: List[Dispersion]

    def __init__(self, *disps: Dispersion) -> None:
        super().__init__()

        self.dispersions = []
        for disp in disps:
            self += disp

    def __add__(self, other: Union[int, float, "Dispersion"]) -> "DispersionSum":
        self._check_valid_operand(other)

        if isinstance(other, DispersionSum):
            self.dispersions += other.dispersions
            return self

        if isinstance(other, (int, float)):
            self.dispersions.append(dispersions.EpsilonInf(eps=other))
            return self

        self.dispersions.append(other)
        return self

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        dielectric_function = sum(
            disp.dielectric_function(lbda) for disp in self.dispersions
        )
        return np.array(dielectric_function)

    def __repr__(self):
        return (
            "DispersionSum\n"
            + "=" * 13
            + "\n\n"
            + "\n\n".join(map(str, self.dispersions))
        )


class IndexDispersionSum(IndexDispersion):
    """Represents the sum of two index dispersions"""

    single_params_template: dict = {}
    rep_params_template: dict = {}
    index_dispersions: List[IndexDispersion]

    def __init__(self, *disps: IndexDispersion) -> None:
        super().__init__()

        self.index_dispersions = []
        for disp in disps:
            self += disp

    def __add__(
        self, other: Union[int, float, "IndexDispersion"]
    ) -> "IndexDispersionSum":
        self._check_valid_operand(other)

        if isinstance(other, IndexDispersionSum):
            self.index_dispersions += other.index_dispersions
            return self

        if isinstance(other, (int, float)):
            self.index_dispersions.append(dispersions.ConstantRefractiveIndex(n=other))
            return self

        self.index_dispersions.append(other)
        return self

    def refractive_index(self, lbda: npt.ArrayLike) -> npt.NDArray:
        refractive_index = sum(
            disp.refractive_index(lbda) for disp in self.index_dispersions
        )
        return np.array(refractive_index)

    def __repr__(self):
        return (
            "IndexDispersionSum\n"
            + "=" * 13
            + "\n\n"
            + "\n\n".join(map(str, self.index_dispersions))
        )



# Encoding: utf-8
"""Dispersion specified by a table of wavelengths (nm) and refractive index values."""

from typing import Union
import numpy.typing as npt
import scipy.interpolate

from elli.dispersions.constant_refractive_index import ConstantRefractiveIndex

from .base_dispersion import IndexDispersion, IndexDispersionSum, InvalidParameters


class Table(IndexDispersion):
    """Dispersion specified by a table of wavelengths (nm) and refractive index values.
    Please not that this model will produce errors for wavelengths outside the provided
    wavelength range.

    Single parameters:
        :lbda (list): Wavelengths in nm. This value must be provided.
        :n: Complex refractive index values in the convention n + ik.
            This value must be provided.
        :kind: Type of interpolation
            (see scipy.interpolate.interp1d for more information). Defaults to 'linear'.

    Repeated parameters:
        --

    Output:
        The interpolation in the given wavelength range.
    """

    single_params_template = {"lbda": None, "n": None}
    rep_params_template = {}

    def __init__(self, *args, **kwargs) -> None:
        self.kind = kwargs.pop("kind", "linear")

        super().__init__(*args, **kwargs)

        if len(self.single_params.get("lbda")) == 0:
            raise InvalidParameters("Wavelength array cannot be of length zero.")

        if len(self.single_params.get("n")) != len(self.single_params.get("lbda")):
            raise InvalidParameters(
                "Wavelength and refractive index arrays must have the same length."
            )

        self.interpolation = scipy.interpolate.interp1d(
            self.single_params.get("lbda"),
            self.single_params.get("n"),
            kind=self.kind,
        )

        self.default_lbda_range = self.single_params.get("lbda")

    def __add__(
        self, other: Union[int, float, "IndexDispersion"]
    ) -> "IndexDispersionSum":
        if isinstance(other, (int, float)):
            return IndexDispersionSum(self, ConstantRefractiveIndex(eps=other))

        if isinstance(other, Table):
            raise NotImplementedError(
                "Adding of tabular dispersions is not yet supported"
            )

        if isinstance(other, IndexDispersion):
            return IndexDispersionSum(self, other)

        if isinstance(other, IndexDispersionSum):
            other.dispersions.append(self)
            return other

        raise TypeError(
            f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'"
        )

    def refractive_index(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return self.interpolation(lbda)



# Encoding: utf-8
"""Lorentz dispersion law with parameters in units of wavelengths."""

import numpy.typing as npt

from .base_dispersion import Dispersion


class LorentzLambda(Dispersion):
    r"""Lorentz dispersion law with parameters in units of wavelengths.

    Single parameters:
        --

    Repeated parameters:
        :A: Amplitude of the oscillator. Defaults to 1.
        :lambda_r: Resonance wavelength. Defaults to 0. Unit in nm.
        :gamma: Broadening of the oscillator. Defaults to 0. Unit in nm.

    Output:

        .. math::
            \varepsilon(\lambda) = 1 + \sum_j \boldsymbol{A}_j
            \cdot \lambda^2 / (\lambda^2 - \boldsymbol{lambda\_r}_j^2
            + i \cdot \boldsymbol{gamma}_j \cdot \lambda)

        The summation index :math:`j` refers to the respective oscillator.
    """

    single_params_template = {}
    rep_params_template = {"A": 1, "lambda_r": 0, "gamma": 0}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return 1 + sum(
            c.get("A")
            * lbda**2
            / (lbda**2 - c.get("lambda_r") ** 2 - 1j * c.get("gamma") * lbda)
            for c in self.rep_params
        )



# Encoding: utf-8
"""Cauchy dispersion, with Urbach tail."""

import numpy as np
import numpy.typing as npt

from .base_dispersion import IndexDispersion
from ..utils import conversion_wavelength_energy


class CauchyUrbach(IndexDispersion):
    r"""Cauchy dispersion, with an Urbach Tail absorption.

    Single parameters:
        :n0: Defaults to 1.5.
        :B: Defaults to 0. Unit in 1/eV\ :sup:`2`.
        :C: Defaults to 0. Unit in 1/eV\ :sup:`4`.
        :D: Defaults to 0.
        :Eg: Defaults to 0. Unit in eV.
        :Eu: Defaults to 1. Unit in eV.

    Repeated parameters:
        --

    Output:
        .. math::
            n(E) =
            \boldsymbol{n_0} + \boldsymbol{B} E^2 + \boldsymbol{C} E^4
            + i \boldsymbol{D} \exp (\frac{E - \boldsymbol{E_g}}{\boldsymbol{E_u}})

    References:
        * Fujiwara: Spectroscopic Ellipsometry: Principles and Applications,
          John Wiley & Sons Ltd, 2007, p. 258
    """

    single_params_template = {"n0": 1.5, "B": 0, "C": 0, "D": 0, "Eg": 2, "Eu": 0.5}
    rep_params_template = {}

    def refractive_index(self, lbda: npt.ArrayLike) -> npt.NDArray:
        energy = conversion_wavelength_energy(lbda)
        return (
            self.single_params.get("n0")
            + self.single_params.get("B") * energy**2
            + self.single_params.get("C") * energy**4
            + 1j
            * self.single_params.get("D")
            * np.exp(
                (energy - self.single_params.get("Eg")) / self.single_params.get("Eu")
            )
        )



# Encoding: utf-8
"""Sellmeier dispersion."""

import numpy.typing as npt

from .base_dispersion import Dispersion


class Sellmeier(Dispersion):
    r"""Sellmeier dispersion.

    Single parameters:
        --

    Repeated parameters:
        :A: Coefficient for n\ :sup:`2` contribution. Defaults to 0.
        :B: Resonance wavelength. Defaults to 0. Unit in µm\ :sup:`2`.

    Output:
        .. math::
            \varepsilon(\lambda) = 1 + \sum_j \boldsymbol{A}_j
            \cdot \lambda^2 /(\lambda^2 - \boldsymbol{B}_j)

        With :math:`j` as the index of the respective oscillator.
    """

    single_params_template = {}
    rep_params_template = {"A": 0, "B": 0}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        lbda = lbda / 1e3
        return 1 + sum(
            c.get("A") * lbda**2 / (lbda**2 - c.get("B")) for c in self.rep_params
        )



# Encoding: utf-8
"""Helper class to load spectraray's tabulated dielectric functions."""

import pandas as pd

from ..utils import conversion_wavelength_energy
from .table_epsilon import TableEpsilon


class TableSpectraRay:
    """Helper class to load spectraray's tabulated dielectric functions."""

    def __init__(self, path: str) -> None:
        """
        Args:
            path (str): Defines the folder where the Spectraray files are saved.
        """
        self.spectraray_path = path

    def load_dispersion_table(self, fname: str) -> TableEpsilon:
        """Load a dispersion table from a ascii file
        in the spectraray materials ascii format.
        This only accounts for tabulated dielectric function data.
        Spectraray also stores dispersion data in other formats,
        but this function is not able to read these other formats.

        Args:
            fname (str): The filename of the spectraray ascii file.

        Returns:
            TableEpsilon: A dispersion object containing the tabulated data.
        """
        start = 0
        stop = 0
        with open(self.spectraray_path + fname, "r", encoding="utf8") as file:
            line = file.readline()
            cnt = 0
            while line:
                if line.strip() == "Begin of array":
                    start = cnt + 1
                if line.strip() == "End of array":
                    stop = cnt
                line = file.readline()
                cnt += 1

                if line.startswith("Units="):
                    x_unit = line.split("=")[1].split(",")[0]

        df = pd.read_csv(
            self.spectraray_path + fname,
            sep=r"\s+",
            skiprows=start,
            nrows=stop - start,
            index_col=0,
            usecols=[0, 1, 2],
            names=[x_unit, "ϵ1", "ϵ2"],
        )

        if x_unit == "Wavelength":
            return TableEpsilon(
                lbda=df.index, epsilon=df.loc[:, "ϵ1"] + 1j * df.iloc[:, "ϵ2"]
            )
        return TableEpsilon(
            lbda=conversion_wavelength_energy(df.index),
            epsilon=df.loc[:, "ϵ1"] + 1j * df.loc[:, "ϵ2"],
        )



#!/usr/bin/python
# encoding: utf-8

import elli
import numpy as np
from scipy.constants import pi


class TestTirThickness:
    # Refractive indices
    n_f = 1.5
    n_s = 1.0
    n_b = 1.7

    # Materials:
    glass1 = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_f))
    air = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_s))
    glass2 = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_b))

    # Air layer thickness variation range
    d = np.linspace(0, 1000)

    # Structure:
    structures = []

    for dd in d:
        layer = elli.Layer(air, dd)
        structures.append(elli.Structure(glass1, [layer], glass2))

    # Wavelength and wavenumber:
    lbda = 1000
    k0 = 2 * pi / lbda
    Phi_i = pi / 2 * 0.6  #  Incidence angle (higher than the limit angle)

    # Analytical calculation

    # Reduced wavenumber
    Kx = n_f * np.sin(Phi_i)

    # Incidence angle
    Phi_s = np.arcsin((complex(Kx / n_s)))
    Phi_b = np.arcsin(Kx / n_b)

    # Wave vector:
    kz_f = n_f * k0 * np.cos(Phi_i)
    kz_s = k0 * np.sqrt(complex(n_s**2 - Kx**2))
    kz_b = n_b * k0 * np.cos(Phi_b)

    # Amplitude coefficient polarisation s:
    r_sf_s = (kz_f - kz_s) / (kz_s + kz_f)
    r_bs_s = (kz_s - kz_b) / (kz_s + kz_b)
    t_sf_s = 1 + r_sf_s
    t_bs_s = 1 + r_bs_s

    # Amplitude coefficient polarisation p:
    r_sf_p = (kz_f * n_s**2 - kz_s * n_f**2) / (kz_s * n_f**2 + kz_f * n_s**2)
    r_bs_p = (kz_s * n_b**2 - kz_b * n_s**2) / (kz_s * n_b**2 + kz_b * n_s**2)
    t_sf_p = np.cos(Phi_i) * (1 - r_sf_p) / np.cos(Phi_s)
    t_bs_p = np.cos(Phi_s) * (1 - r_bs_p) / np.cos(Phi_b)

    # Power coefficients:
    R_th_s = (
        np.abs(
            (r_sf_s + r_bs_s * np.exp(2j * kz_s * d))
            / (1 + r_bs_s * r_sf_s * np.exp(2j * kz_s * d))
        )
    ) ** 2

    t2_th_s = (
        np.abs(
            (t_bs_s * t_sf_s * np.exp(1j * kz_s * d))
            / (1 + r_bs_s * r_sf_s * np.exp(2j * kz_s * d))
        )
    ) ** 2

    R_th_p = (
        np.abs(
            (r_sf_p + r_bs_p * np.exp(2j * kz_s * d))
            / (1 + r_bs_p * r_sf_p * np.exp(2j * kz_s * d))
        )
    ) ** 2

    t2_th_p = (
        np.abs(
            (t_bs_p * t_sf_p * np.exp(1j * kz_s * d))
            / (1 + r_bs_p * r_sf_p * np.exp(2j * kz_s * d))
        )
    ) ** 2

    correction = np.real(n_b * np.cos(Phi_b) / (n_f * np.cos(Phi_i)))
    # This is a correction term used in R +T*correction = 1

    T_th_s = t2_th_s * correction
    T_th_p = t2_th_p * correction

    def test_tir_thickness_4x4_expm(self):
        data = elli.ResultList(
            [s.evaluate(self.lbda, np.rad2deg(self.Phi_i)) for s in self.structures]
        )

        # Extraction of the transmission and reflexion coefficients
        R_p = data.R_pp
        R_s = data.R_ss
        T_p = data.T_pp
        T_s = data.T_ss
        t2_p = np.abs(data.t_pp) ** 2  # Before power correction
        t2_s = np.abs(data.t_ss) ** 2

        np.testing.assert_allclose(R_p, self.R_th_p)
        np.testing.assert_allclose(R_s, self.R_th_s)
        np.testing.assert_allclose(T_p, self.T_th_p)
        np.testing.assert_allclose(T_s, self.T_th_s)
        np.testing.assert_allclose(t2_p, self.t2_th_p)
        np.testing.assert_allclose(t2_s, self.t2_th_s)

    def test_tir_thickness_4x4_eig(self):
        data = elli.ResultList(
            [
                s.evaluate(
                    self.lbda,
                    np.rad2deg(self.Phi_i),
                    solver=elli.Solver4x4,
                    propagator=elli.PropagatorEig(),
                )
                for s in self.structures
            ]
        )

        # Extraction of the transmission and reflexion coefficients
        R_p = data.R_pp
        R_s = data.R_ss
        T_p = data.T_pp
        T_s = data.T_ss
        t2_p = np.abs(data.t_pp) ** 2  # Before power correction
        t2_s = np.abs(data.t_ss) ** 2

        np.testing.assert_allclose(R_p, self.R_th_p)
        np.testing.assert_allclose(R_s, self.R_th_s)
        np.testing.assert_allclose(T_p, self.T_th_p)
        np.testing.assert_allclose(T_s, self.T_th_s)
        np.testing.assert_allclose(t2_p, self.t2_th_p)
        np.testing.assert_allclose(t2_s, self.t2_th_s)

    def test_tir_thickness_2x2(self):
        data = elli.ResultList(
            [
                s.evaluate(self.lbda, np.rad2deg(self.Phi_i), solver=elli.Solver2x2)
                for s in self.structures
            ]
        )

        # Extraction of the transmission and reflexion coefficients
        R_p = data.R_pp
        R_s = data.R_ss
        T_p = data.T_pp
        T_s = data.T_ss
        t2_p = np.abs(data.t_pp) ** 2  # Before power correction
        t2_s = np.abs(data.t_ss) ** 2

        np.testing.assert_allclose(R_p, self.R_th_p)
        np.testing.assert_allclose(R_s, self.R_th_s)
        np.testing.assert_allclose(T_p, self.T_th_p)
        np.testing.assert_allclose(T_s, self.T_th_s)
        np.testing.assert_allclose(t2_p, self.t2_th_p)
        np.testing.assert_allclose(t2_s, self.t2_th_s)


class TestTirAngle:
    n_f = 1.5
    n_s = 1.0
    n_b = 1.7

    # Materials:
    glass1 = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_f))
    air = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_s))
    glass2 = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_b))

    # Layer:
    layer = elli.Layer(air, 0)

    # Structure:
    s = elli.Structure(glass1, [layer], glass2)

    # Wavelength and wavenumber:
    lbda = 1000
    k0 = 2 * pi / lbda

    # Layer thickness
    d = lbda * 0.347
    layer.set_thickness(d)

    # Variation of incidence angle
    Phi_list = np.deg2rad(np.linspace(0, 89, 90))

    # Analytical calculation

    # Reduced wavenumber
    Kx = n_f * np.sin(Phi_list)

    Phi_s = np.arcsin((Kx / n_s).astype(complex))
    Phi_b = np.arcsin((Kx / n_b).astype(complex))

    # Wave vector:
    kz_f = n_f * k0 * np.cos((Phi_list.astype(complex)))
    kz_s = k0 * np.sqrt((-(Kx**2 - n_s**2)).astype(complex))
    kz_b = n_b * k0 * np.cos(Phi_b)

    # Amplitude coefficient for 's' polarisation:
    r_sf_s = (kz_f - kz_s) / (kz_s + kz_f)
    r_bs_s = (kz_s - kz_b) / (kz_s + kz_b)
    t_sf_s = 1 + r_sf_s
    t_bs_s = 1 + r_bs_s

    # Amplitude coefficient for 'p' polarisation:
    r_sf_p = (kz_f * n_s**2 - kz_s * n_f**2) / (kz_s * n_f**2 + kz_f * n_s**2)
    r_bs_p = (kz_s * n_b**2 - kz_b * n_s**2) / (kz_s * n_b**2 + kz_b * n_s**2)
    t_sf_p = np.cos((Phi_list.astype(complex))) * (1 - r_sf_p) / np.cos(Phi_s)
    t_bs_p = np.cos(Phi_s) * (1 - r_bs_p) / np.cos(Phi_b)

    # Power coefficients:
    R_th_s = (
        np.abs(
            (r_sf_s + r_bs_s * np.exp(2j * kz_s * d))
            / (1 + r_bs_s * r_sf_s * np.exp(2j * kz_s * d))
        )
    ) ** 2

    t2_th_s = (
        np.abs(
            (t_bs_s * t_sf_s * np.exp(1j * kz_s * d))
            / (1 + r_bs_s * r_sf_s * np.exp(2j * kz_s * d))
        )
    ) ** 2

    R_th_p = (
        np.abs(
            (r_sf_p + r_bs_p * np.exp(2j * kz_s * d))
            / (1 + r_bs_p * r_sf_p * np.exp(2j * kz_s * d))
        )
    ) ** 2

    t2_th_p = (
        np.abs(
            (t_bs_p * t_sf_p * np.exp(1j * kz_s * d))
            / (1 + r_bs_p * r_sf_p * np.exp(2j * kz_s * d))
        )
    ) ** 2

    correction = np.real(n_b * np.cos(Phi_b) / (n_f * np.cos(Phi_list.astype(complex))))
    # This is a correction term used in R +T*correction = 1

    T_th_s = t2_th_s * correction
    T_th_p = t2_th_p * correction

    def test_tir_angle_4x4_expm(self):
        data = elli.ResultList(
            [self.s.evaluate(self.lbda, np.rad2deg(Phi_i)) for Phi_i in self.Phi_list]
        )

        # Extraction of the transmission and reflexion coefficients
        R_p = data.R_pp
        R_s = data.R_ss
        T_p = data.T_pp
        T_s = data.T_ss
        t2_p = np.abs(data.t_pp) ** 2  # Before power correction
        t2_s = np.abs(data.t_ss) ** 2

        np.testing.assert_allclose(R_p, self.R_th_p)
        np.testing.assert_allclose(R_s, self.R_th_s)
        np.testing.assert_allclose(T_p, self.T_th_p)
        np.testing.assert_allclose(T_s, self.T_th_s)
        np.testing.assert_allclose(t2_p, self.t2_th_p)
        np.testing.assert_allclose(t2_s, self.t2_th_s)

    def test_tir_angle_4x4_eig(self):
        data = elli.ResultList(
            [
                self.s.evaluate(
                    self.lbda,
                    np.rad2deg(Phi_i),
                    solver=elli.Solver4x4,
                    propagator=elli.PropagatorEig(),
                )
                for Phi_i in self.Phi_list
            ]
        )

        # Extraction of the transmission and reflexion coefficients
        R_p = data.R_pp
        R_s = data.R_ss
        T_p = data.T_pp
        T_s = data.T_ss
        t2_p = np.abs(data.t_pp) ** 2  # Before power correction
        t2_s = np.abs(data.t_ss) ** 2

        np.testing.assert_allclose(R_p, self.R_th_p)
        np.testing.assert_allclose(R_s, self.R_th_s)
        np.testing.assert_allclose(T_p, self.T_th_p)
        np.testing.assert_allclose(T_s, self.T_th_s)
        np.testing.assert_allclose(t2_p, self.t2_th_p)
        np.testing.assert_allclose(t2_s, self.t2_th_s)

    def test_tir_angle_2x2(self):
        data = elli.ResultList(
            [
                self.s.evaluate(self.lbda, np.rad2deg(Phi_i), solver=elli.Solver2x2)
                for Phi_i in self.Phi_list
            ]
        )

        # Extraction of the transmission and reflexion coefficients
        R_p = data.R_pp
        R_s = data.R_ss
        T_p = data.T_pp
        T_s = data.T_ss
        t2_p = np.abs(data.t_pp) ** 2  # Before power correction
        t2_s = np.abs(data.t_ss) ** 2

        np.testing.assert_allclose(R_p, self.R_th_p)
        np.testing.assert_allclose(R_s, self.R_th_s)
        np.testing.assert_allclose(T_p, self.T_th_p)
        np.testing.assert_allclose(T_s, self.T_th_s)
        np.testing.assert_allclose(t2_p, self.t2_th_p)
        np.testing.assert_allclose(t2_s, self.t2_th_s)


def test_4x4_transition_matrix_halfspace():
    # Refractive indices
    n_glass = 1.5
    n_air = 1.0

    # Materials:
    glass = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_glass))
    air = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_air))

    # Pseudo uniaxial Backhalfspace, to force different algorithm
    air_uniaxial = elli.UniaxialMaterial(
        elli.ConstantRefractiveIndex(n=n_air), elli.ConstantRefractiveIndex(n=n_air)
    )

    # Structure:
    s = elli.Structure(glass, [], air)
    s2 = elli.Structure(glass, [], air_uniaxial)

    # Wavelength
    lbda = 1000  # nm

    # Variation of incidence angle
    Phi_list = np.linspace(0, 89, 90)

    R = []
    R_uni = []

    for Phi_i in Phi_list:
        data = s.evaluate(lbda, Phi_i)
        R.append(data.R[0])

        data2 = s2.evaluate(lbda, Phi_i)
        R_uni.append(data2.R[0])

    R = np.array(R)
    R_uni = np.array(R_uni)

    np.testing.assert_allclose(R, R_uni)



"""A helper class to load data from SpectraRay ASCII Files.
It only supplies a rudimentary loading of standard psi/delta values
and misses some other features.
"""

import re

import pandas as pd
from packaging.version import Version, parse

from ..utils import calc_rho
from . import detect_encoding


def read_spectraray_psi_delta(
    fname: str, sep: str = r"\s+", decimal: str = "."
) -> pd.DataFrame:
    r"""Read a psi/delta spectraray ascii file.

    Args:
        fname (str): Filename of the measurement ascii file.
        sep (str, optional): Data separator in the datafile. Defaults to "\s+".
        decimal (str, optional): Decimal separator in the datafile. Defaults to ".".

    Returns:
        pd.DataFrame: DataFrame containing the psi/delta data in
        the format to be further processes inside pyElli.
    """
    # detect encoding
    encoding = detect_encoding(fname)

    # read data and drop empty column
    psi_delta_df = pd.read_csv(
        fname,
        encoding=encoding,
        index_col=0,
        header=None,
        sep=sep,
        decimal=decimal,
        skiprows=1,
    )
    psi_delta_df.dropna(axis="columns", how="all", inplace=True)

    # index data correctly
    psi_delta_df.index.name = "Wavelength"

    with open(fname) as f:
        header = f.readlines()[0]

    aois = list(map(float, re.split(sep, header)[3::2]))
    index = pd.MultiIndex.from_product(
        [aois, ["Ψ", "Δ"]], names=["Angle of Incidence", ""]
    )
    psi_delta_df.columns = index

    # reorder dataframe
    if Version("2.2") <= parse(pd.__version__) < Version("3.0"):
        psi_delta_df = psi_delta_df.stack(0, future_stack=True)
    else:
        psi_delta_df = psi_delta_df.stack(0)
    psi_delta_df = psi_delta_df.reorder_levels(["Angle of Incidence", "Wavelength"])
    psi_delta_df.sort_index(axis=0, inplace=True)
    psi_delta_df.sort_index(axis=1, ascending=False, inplace=True)

    # convert delta range
    psi_delta_df.loc[:, "Δ"] = psi_delta_df.loc[:, "Δ"].where(
        psi_delta_df.loc[:, "Δ"] <= 180, psi_delta_df.loc[:, "Δ"] - 360
    )

    return psi_delta_df


def read_spectraray_mmatrix(
    fname: str, sep: str = r"\s+", decimal: str = "."
) -> pd.DataFrame:
    r"""Read a mueller matrix spectraray ascii file.
    Only reads the first entry and does not support reading multiple angles.
    For multiple angles you have to save the data in multiple files.

    Args:
        fname (str): Filename of the measurement ascii file.
        sep (str, optional): Data separator in the datafile. Defaults to "\s+".
        decimal (str, optional): Decimal separator in the datafile. Defaults to ".".

    Returns:
        pd.DataFrame: DataFrame containing the psi/delta data in
        the format to be further processes inside pyElli.
    """
    encoding = detect_encoding(fname)

    mueller_matrix = pd.read_csv(
        fname, encoding=encoding, sep=sep, decimal=decimal, index_col=0
    ).iloc[:, -17:-1]
    mueller_matrix.index.name = "Wavelength"
    mueller_matrix.columns = [
        "M11",
        "M12",
        "M13",
        "M14",
        "M21",
        "M22",
        "M23",
        "M24",
        "M31",
        "M32",
        "M33",
        "M34",
        "M41",
        "M42",
        "M43",
        "M44",
    ]

    return mueller_matrix


def read_spectraray_rho(
    fname: str, sep: str = r"\s+", decimal: str = "."
) -> pd.DataFrame:
    r"""Read a psi/delta spectraray ascii file and converts it to rho values.

    Args:
        fname (str): Filename of the measurement ascii file.
        sep (str, optional): Data separator in the datafile. Defaults to "\s+".
        decimal (str, optional): Decimal separator in the datafile. Defaults to ".".

    Returns:
        pd.DataFrame: DataFrame containing the rho data in
        the format to be further processes inside pyElli.
    """
    psi_delta = read_spectraray_psi_delta(fname, sep, decimal)
    return calc_rho(psi_delta)



"""Plotting functions for Mueller matrices"""

# Encoding: utf-8
from typing import List, Union, Optional

import pandas as pd

try:
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
except ImportError as e:
    raise ImportError(
        "Optional dependency plotly is not installed. This module will not work properly.\n"
        "Try installing this package with the additional fitting requirement, "
        "i.e. pip install pyElli[fitting]"
    ) from e

COLORS = [
    "#636EFA",
    "#EF553B",
    "#00CC96",
    "#AB63FA",
    "#FFA15A",
    "#19D3F3",
    "#FF6692",
    "#B6E880",
    "#FF97FF",
    "#FECB52",
]


def plot_mmatrix(
    dataframes: Union[pd.DataFrame, List[pd.DataFrame]],
    colors: Optional[List[str]] = None,
    dashes: Optional[List[str]] = None,
    names: Optional[List[str]] = None,
    single: bool = True,
    full_scale: bool = False,
    sharex: bool = False,
) -> go.Figure:
    """Takes multiple Mueller matrix dataframes with columns Mxy for matrix postion x,y
    and plots them together. Needs plotly as additional requirement to work.

    Args:
        dataframes (Union[pd.Dataframe, List[pd.DataFrame]]):
            A dataframe or a list of dataframes containing data of the same index.
        colors (Optional[List[str]], optional):
            A list of colors which are cycled for each dataframes index. Defaults to None.
        dashes (Optional[List[str]], optional):
            A list of dash line styles which are cycled for each dataframes index. Defaults to None.
        names (Optional[List[str]], optional): A name for each dataframe index. Defaults to None.
        single (bool, optional): Uses a single plot if set and a grid if not set. Defaults to True.
        full_scale (bool, optional): Sets the y-axis limits to [-1, 1] if set. Defaults to False.
        sharex (bool, optional):
            Ties the zooming of the x-axis together for each plot in grid view. Defaults to False.

    Returns:
        go.Figure: A plotly figure containing the data from dataframes as a grid or single view.
    """
    if isinstance(dataframes, pd.DataFrame):
        dataframes = [dataframes]

    if colors is None:
        colors = COLORS
    if dashes is None:
        dashes = ["solid", "dash", "dot"]
    if names is None:
        names = ["", "theory"]

    if single:
        fig = go.Figure()
        fig.update_layout(
            yaxis_title="Müller Matrix Elements", xaxis_title="Wavelength (nm)"
        )
    else:
        fig = make_subplots(rows=4, cols=4)
        if full_scale:
            fig.update_yaxes(range=[-1, 1])

    for i, melem in enumerate(dataframes[0]):
        coli = colors[i % len(colors)]
        for j, mueller_df in enumerate(dataframes):
            dashi = dashes[j % len(dashes)]
            namesi = names[j % len(names)]
            if single:
                fig.add_trace(
                    go.Scatter(
                        x=mueller_df.index,
                        y=mueller_df[melem],
                        name=f"{melem} {namesi}",
                        line=dict(color=coli, dash=dashi),
                    )
                )
            else:
                fig.add_trace(
                    go.Scatter(
                        x=mueller_df.index,
                        y=mueller_df[melem],
                        name=f"{melem} {namesi}",
                        line=dict(color=coli, dash=dashi),
                    ),
                    row=1 if single else i // 4 + 1,
                    col=1 if single else i % 4 + 1,
                )
    if sharex:
        fig.update_xaxes(matches="x")
    return go.FigureWidget(fig)



# Encoding: utf-8
"""Helper class to use the refractiveindex.info database."""

import io
import os
import re
from collections import namedtuple
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
import yaml
from importlib_resources import files
from rapidfuzz import process

from ..dispersions import (
    CauchyCustomExponent,
    EpsilonInf,
    Polynomial,
    Sellmeier,
    SellmeierCustomExponent,
    Table,
)
from ..dispersions.base_dispersion import (
    Dispersion,
    DispersionSum,
    IndexDispersion,
    IndexDispersionSum,
)
from ..materials import IsotropicMaterial

WavelengthFilterType = Union[
    None, float, int, List[Union[float, int]], Tuple[Union[float, int]]
]

nt_entry = namedtuple(
    "Entry",
    [
        "shelf",
        "shelf_longname",
        "book_divider",
        "book",
        "book_longname",
        "page",
        "page_type",
        "author",
        "year",
        "comment",
        "lower_range",
        "upper_range",
        "path",
    ],
)


class RII:
    """Helper class to load tabulated dielectric functions from the https://refractiveindex.info database.

    The database object has to be initialized to load the entire catalog:

    .. highlight:: python
    .. code-block:: python

        RII = elli.db.RII()

    After initialization the Class provides a dataframe with all entries at RII.catalog,
    it follows the naming schema used on the website: Entries are categorized by Shelf, Book and Page.

    Shelfs are broad categories (inorganics, organics, glasses, other), and ignored in the following.

    Book and Page are used to access entries with this helper:
    For inorganic and organic materials 'book' uses the sum formula of the compound and
    'page' is an identifier of the author of the publication.
    In the case of glasses, book is the manufacturer and page the name of the glass.

    Materials can be searched by the RII.search() method.
    Specific fields can be searched by providing a column keyword:

    .. highlight:: python
    .. code-block:: python

        RII.search("Aspnes", column="author")

    Dispersions or respective materials can be loaded by calling these methods:

    .. highlight:: python
    .. code-block:: python

        gold_material = RII.get_mat("Au", "Johnson")
        gold_dispersion = RII.get_dispersion("Au", "Johnson")
    """

    def __init__(self) -> None:
        self.rii_path = files("elli.database.refractiveindexinfo-database.database")

        with open(self.rii_path.joinpath("catalog-nk.yml"), "r", encoding="utf8") as f:
            yml_file = yaml.load(f, yaml.SafeLoader)

        pagename_pattern = re.compile(
            r"(?P<authors>.*) "
            r"(?P<year>\d{4})[^:]*?: "
            r"((?P<comment1>.*); )?"
            r"(?P<type1>.+) (?P<lower_range1>\d+(\.\d*)?(e\W?\d+)?)\W(?P<upper_range1>\d+(\.\d*)?(e\W?\d+)?) µm"
            r"(, (?P<type2>.+) (?P<lower_range2>\d+(\.\d*)?(e\W?\d+)?).(?P<upper_range2>\d+(\.\d*)?(e\W?\d+)?) µm)?"
            r"(; (?P<comment2>.*))?"
        )

        entries = []
        for sh in yml_file:
            b_div = pd.NA
            for b in sh["content"]:
                if "DIVIDER" not in b:
                    p_div = pd.NA
                    for p in b["content"]:
                        if "DIVIDER" not in p:
                            infos = pagename_pattern.match(p["name"])
                            if infos is None:
                                entries.append(
                                    nt_entry(
                                        sh["SHELF"],
                                        sh["name"],
                                        b_div,
                                        b["BOOK"],
                                        b["name"],
                                        p["PAGE"],
                                        p_div,
                                        None,
                                        None,
                                        p["name"],
                                        None,
                                        None,
                                        os.path.join(
                                            "data-nk", os.path.normpath(p["data"])
                                        ),
                                    )
                                )
                            else:
                                entries.append(
                                    nt_entry(
                                        sh["SHELF"],
                                        sh["name"],
                                        b_div,
                                        b["BOOK"],
                                        b["name"],
                                        p["PAGE"],
                                        p_div,
                                        infos.group("authors"),
                                        infos.group("year"),
                                        " ".join(
                                            filter(
                                                None,
                                                (
                                                    infos.group("comment1"),
                                                    infos.group("comment2"),
                                                ),
                                            )
                                        ),
                                        infos.group("lower_range1"),
                                        infos.group("upper_range1"),
                                        os.path.join(
                                            "data-nk", os.path.normpath(p["data"])
                                        ),
                                    )
                                )
                        else:
                            p_div = p["DIVIDER"]
                else:
                    b_div = b["DIVIDER"]

        self.catalog = pd.DataFrame(entries, dtype=pd.StringDtype())

        self.catalog["year"] = pd.to_numeric(
            self.catalog["year"], errors="coerce"
        ).convert_dtypes()
        self.catalog["lower_range"] = 1000 * pd.to_numeric(
            self.catalog["lower_range"], errors="coerce"
        )
        self.catalog["upper_range"] = 1000 * pd.to_numeric(
            self.catalog["upper_range"], errors="coerce"
        )

        book_div = self.catalog["book_divider"].drop_duplicates().dropna().values
        books = self.catalog["book"].drop_duplicates().dropna().values
        book_longnames = self.catalog["book_longname"].drop_duplicates().dropna().values
        pages = self.catalog["page"].drop_duplicates().dropna().values
        authors = self.catalog["author"].drop_duplicates().dropna().values
        comments = self.catalog["comment"].drop_duplicates().dropna().values

        self._filter_lists = {
            "book_divider": book_div,
            "book": books,
            "book_longname": book_longnames,
            "page": pages,
            "author": authors,
            "comment": comments,
        }

    def search(
        self,
        query: Union[str, List[str]],
        column: Union[str, List[str]] = "all",
        wavelength_filter: WavelengthFilterType = None,
        fuzzy: bool = True,
    ) -> pd.DataFrame:
        """Search the catalog by the query string in the requested column.

        Args:
            query (str, List[str]): String or list of strings to search.
            column (str, optional): Column-strings or list of strings to search.
            wavelength_filter (float, int, List[float, int]): Wavelengths in nm included in the results. Default to None.
            fuzzy (bool, optional): Search approximate entries. Defaults to True.

        Returns:
            pd.DataFrame: Filtered Catalog dataframe.
        """
        if isinstance(query, str):
            query = [query]

        if column == "all":
            if fuzzy:
                column = list(self._filter_lists.keys())
            else:
                column = self.catalog.columns
        elif isinstance(column, str):
            column = [column]

        index = self.catalog["shelf"] == ""

        for col in column:
            for q in query:
                if fuzzy:
                    suggestions = process.extract(
                        q, self._filter_lists[col], limit=10, score_cutoff=80
                    )

                    for s, _, _ in suggestions:
                        index = np.logical_or(index, self.catalog[col] == s)

                else:
                    index = np.logical_or(index, self.catalog[col] == q)

        result = self.catalog.loc[index]

        if wavelength_filter is None:
            return result
        elif isinstance(wavelength_filter, (int, float)):
            return result.loc[
                (result.lower_range <= wavelength_filter)
                & (result.upper_range >= wavelength_filter)
            ]
        elif isinstance(wavelength_filter, (list, tuple, np.ndarray)):
            for wl in wavelength_filter:
                result = result.loc[
                    (result["lower_range"] <= wl) & (result["upper_range"] >= wl)
                ]
            return result
        else:
            raise (
                ValueError(
                    "Wavelength_filter only takes numeric values or a list of numeric values."
                )
            )

    def get_mat(self, book: str, page: str) -> IsotropicMaterial:
        """Load a dispersion from the refractive index database and generates an isotropic material.
        Selection by material and source identifiers.

        Args:
            book (str): Name of the Material, named 'Book' on the website and the database. E.g. 'Au'
            page (str): Name of the Source, named 'Page' on the website and the database. E.g. 'Johnson'

        Returns:
            IsotropicMaterial: A material object build from the tabulated dispersion data.
        """
        return IsotropicMaterial(self.get_dispersion(book, page))

    def get_dispersion(
        self, book: str, page: str
    ) -> Union[Dispersion, IndexDispersion]:
        """Load a dispersion from the refractive index database.
        Selection by material and source identifiers.

        Args:
            book (str): Name of the Material, named 'Book' on the website and the database. E.g. 'Au'
            page (str): Name of the Source, named 'Page' on the website and the database. E.g. 'Johnson'

        Returns:
            Dispersion: A dispersion object containing the tabulated data.
        """

        index = self.catalog.loc[
            (self.catalog["book"] == book) & (self.catalog["page"] == page)
        ].index

        if len(index) != 1:
            raise ValueError("No entry found.")

        yml_file = yaml.load(
            self.rii_path.joinpath(self.catalog.loc[index[0]]["path"]).read_text(),
            yaml.SafeLoader,
        )

        dispersion_list = []
        contains_index_dispersion = False

        for dispersion_relation in yml_file["DATA"]:
            if dispersion_relation["type"] == "tabulated nk":
                df = pd.read_table(
                    io.StringIO(dispersion_relation["data"]),
                    sep="\\s+",
                    names=["Wavelength", "n", "k"],
                )
                df["Wavelength"] = df["Wavelength"] * 1000
                df.set_index("Wavelength", inplace=True)

                dispersion = Table(lbda=df.index, n=df["n"] + 1j * df["k"])
                contains_index_dispersion = True

            elif dispersion_relation["type"] == "tabulated n":
                df = pd.read_table(
                    io.StringIO(dispersion_relation["data"]),
                    sep="\\s+",
                    names=["Wavelength", "n"],
                )
                df["Wavelength"] = df["Wavelength"] * 1000
                df.set_index("Wavelength", inplace=True)

                dispersion = Table(lbda=df.index, n=df["n"])
                contains_index_dispersion = True

            elif dispersion_relation["type"] == "tabulated k":
                df = pd.read_table(
                    io.StringIO(dispersion_relation["data"]),
                    sep="\\s+",
                    names=["Wavelength", "k"],
                )
                df["Wavelength"] = df["Wavelength"] * 1000
                df.set_index("Wavelength", inplace=True)

                dispersion = Table(lbda=df.index, n=0 + 1j * df["k"])
                contains_index_dispersion = True

            elif dispersion_relation["type"] == "formula 1":
                coeffs = list(
                    map(float, dispersion_relation["coefficients"].split(" "))
                )
                a = coeffs[slice(1, len(coeffs), 2)]
                b = coeffs[slice(2, len(coeffs), 2)]

                sell = Sellmeier()
                for a_i, b_i in zip(a, b):
                    sell.add(a_i, b_i**2)

                dispersion = sell + EpsilonInf(coeffs[0])

            elif dispersion_relation["type"] == "formula 2":
                coeffs = list(
                    map(float, dispersion_relation["coefficients"].split(" "))
                )
                a = coeffs[slice(1, len(coeffs), 2)]
                b = coeffs[slice(2, len(coeffs), 2)]

                sell = Sellmeier()
                for a_i, b_i in zip(a, b):
                    sell.add(a_i, b_i)

                dispersion = sell + EpsilonInf(coeffs[0])

            elif dispersion_relation["type"] == "formula 3":
                coeffs = list(
                    map(float, dispersion_relation["coefficients"].split(" "))
                )
                f = coeffs[slice(1, len(coeffs), 2)]
                e = coeffs[slice(2, len(coeffs), 2)]

                poly = Polynomial(coeffs[0])
                for f_i, e_i in zip(f, e):
                    poly.add(f_i / 1e3**e_i, e_i)

                dispersion = poly

            elif dispersion_relation["type"] == "formula 4":
                coeffs = list(
                    map(float, dispersion_relation["coefficients"].split(" "))
                )
                a = coeffs[slice(1, 6, 4)]
                e1 = coeffs[slice(2, 7, 4)]
                b = coeffs[slice(3, 8, 4)]
                e2 = coeffs[slice(4, 9, 4)]
                f = coeffs[slice(9, len(coeffs), 2)]
                e = coeffs[slice(10, len(coeffs), 2)]

                poly = Polynomial(coeffs[0])
                sell = SellmeierCustomExponent()

                for a_i, e1_i, b_i, e2_i in zip(a, e1, b, e2):
                    sell.add(a_i, e1_i, b_i, e2_i)

                for f_i, e_i in zip(f, e):
                    poly.add(f_i / 1e3**e_i, e_i)

                dispersion = poly + sell

            elif dispersion_relation["type"] == "formula 5":
                coeffs = list(
                    map(float, dispersion_relation["coefficients"].split(" "))
                )
                f = coeffs[slice(1, len(coeffs), 2)]
                e = coeffs[slice(2, len(coeffs), 2)]

                cauchy = CauchyCustomExponent(coeffs[0])
                for f_i, e_i in zip(f, e):
                    cauchy.add(f_i / 1e3**e_i, e_i)

                dispersion = cauchy
                contains_index_dispersion = True

            else:
                raise ValueError("Unimplemented Format.")

            dispersion_list.append(dispersion)

        if len(dispersion_list) == 1:
            return dispersion_list[0]

        if contains_index_dispersion:
            for i, dispersion in enumerate(dispersion_list):
                if not isinstance(dispersion, IndexDispersion):
                    dispersion_list[i] = dispersion.as_index()

            return IndexDispersionSum(*dispersion_list)
        return DispersionSum(*dispersion_list)

    def get_reference(self, book: str, page: str) -> str:
        """Reads the reference information from the selected dispersion.

        Args:
            book (str): Name of the Material, named 'Book' on the website and the database. E.g. 'Au'
            page (str): Name of the Source, named 'Page' on the website and the database. E.g. 'Johnson'

        Returns:
            str: Reference information.
        """

        index = self.catalog.loc[
            (self.catalog["book"] == book) & (self.catalog["page"] == page)
        ].index

        if len(index) != 1:
            raise ValueError("No entry found.")

        yml_file = yaml.load(
            self.rii_path.joinpath(self.catalog.loc[index[0]]["path"]).read_text(),
            yaml.SafeLoader,
        )

        return yml_file["REFERENCES"]

    def get_comment(self, book: str, page: str) -> str:
        """Reads the measurement/calculation information of the selected dispersion.

        Args:
            book (str): Name of the Material, named 'Book' on the website and the database. E.g. 'Au'
            page (str): Name of the Source, named 'Page' on the website and the database. E.g. 'Johnson'

        Returns:
            str: Dispersion information.
        """

        index = self.catalog.loc[
            (self.catalog["book"] == book) & (self.catalog["page"] == page)
        ].index

        if len(index) != 1:
            raise ValueError("No entry found.")

        yml_file = yaml.load(
            self.rii_path.joinpath(self.catalog.loc[index[0]]["path"]).read_text(),
            yaml.SafeLoader,
        )

        return yml_file["COMMENTS"]



#!/usr/bin/python
# encoding: utf-8

import elli
import numpy as np
from scipy.constants import pi


class TestBragg:
    n_a = 1.0
    n_g = 1.5
    air = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_a))
    glass = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_g))

    lbda0 = 1550  # nm
    k0 = 2 * pi / lbda0
    n_SiO2 = 1.47
    n_TiO2 = 2.23 + 1j * 5.2e-4

    SiO2 = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_SiO2))
    TiO2 = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_TiO2))

    # Layers and Structure
    d_SiO2 = elli.get_qwp_thickness(SiO2, lbda0)
    d_TiO2 = elli.get_qwp_thickness(TiO2, lbda0)

    L_SiO2 = elli.Layer(SiO2, d_SiO2)
    L_TiO2 = elli.Layer(TiO2, d_TiO2)

    # Repeated layers: n periods
    Layerstack = elli.RepeatedLayers([L_TiO2, L_SiO2], 4, 0, 0)

    # Number of interfaces
    N = 2 * Layerstack.repetitions + 1

    # Structure
    s = elli.Structure(air, [Layerstack], glass)

    # Analytical calculation
    n = np.ones(N + 1, dtype=complex)
    n[0] = n_a
    n[1::2] = n_TiO2
    n[2::2] = n_SiO2
    n[-1] = n_g

    n.shape = (-1, 1)

    d = np.ones(N + 1)
    d[1::2] = L_TiO2.thickness  #  d[0] is not used
    d[2::2] = L_SiO2.thickness

    (lbda1, lbda2) = (1100, 2500)
    lbda_list = np.linspace(lbda1, lbda2, 200)

    def ReflectionCoeff(self, incidence_angle=0.0, polarisation="s"):
        """Returns the reflection coefficient in amplitude"""
        Kx = self.n[0] * np.sin(incidence_angle)
        sinPhi = Kx / self.n
        kz = 2 * pi / self.lbda_list * np.sqrt(self.n**2 - Kx**2)

        # Reflexion coefficient r_{k,k+1} for a single interface
        #    polarisation s:
        #    r_ab(p) = r_{p,p+1} = (kz(p)-kz(p+1))/(kz(p)+kz(p+1))
        #    polarisation p:
        #    r_ab(p) = r_{p,p+1} = (kz(p)*n[p+1]**2-kz(p+1)*n[p]**2) \
        #                          /(kz(p)*n[p]**2+kz(p+1)*n[p+1]**2)
        if polarisation == "s":
            r_ab = (-np.diff(kz, axis=0)) / (kz[:-1] + kz[1:])
        elif polarisation == "p":
            r_ab = (kz[:-1] * (self.n[1:]) ** 2 - kz[1:] * (self.n[:-1]) ** 2) / (
                kz[:-1] * (self.n[1:]) ** 2 + kz[1:] * (self.n[:-1]) ** 2
            )

        # Local function definition for recursive calculation
        def U(k):
            """Returns reflection coefficient U(k) = r_{k, {k+1,...,N}}

            Used recursively.
            """
            p = k + 1
            if p == self.N:
                res = r_ab[self.N - 1]
            else:
                res = (r_ab[p - 1] + U(p) * np.exp(2j * kz[p] * self.d[p])) / (
                    1 + r_ab[p - 1] * U(p) * np.exp(2j * kz[p] * self.d[p])
                )
            return res

        return U(0)

    def test_normal_incidence(self):
        data = self.s.evaluate(self.lbda_list, 0)
        R_ss_0 = data.R_ss
        R_th_ss_0 = (np.abs(self.ReflectionCoeff(0, "s"))) ** 2  # Phi_i = 0
        np.testing.assert_allclose(R_ss_0, R_th_ss_0, 1e-10, 0)

    def test_angle(self):
        R_th_ss = (np.abs(self.ReflectionCoeff(pi / 4, "s"))) ** 2  # Phi_i = pi/4
        R_th_pp = (np.abs(self.ReflectionCoeff(pi / 4, "p"))) ** 2

        # Incidence angle Phi_i = 0, 's' polarization

        # Incidence angle Phi_i = pi/4, 's' and 'p' polarizations
        data2 = self.s.evaluate(self.lbda_list, np.rad2deg(pi / 4))

        R_ss = data2.R_ss
        R_pp = data2.R_pp

        np.testing.assert_array_almost_equal(R_ss, R_th_ss, decimal=1)
        np.testing.assert_array_almost_equal(R_pp, R_th_pp, decimal=1)

    def test_normal_incidence_4x4_expm(self):
        data = self.s.evaluate(self.lbda_list, 0)
        R_ss_0 = data.R_ss
        R_th_ss_0 = (np.abs(self.ReflectionCoeff(0, "s"))) ** 2  # Phi_i = 0
        np.testing.assert_allclose(R_ss_0, R_th_ss_0, 1e-10, 0)

    def test_angle_4x4_expm(self):
        R_th_ss = (np.abs(self.ReflectionCoeff(pi / 4, "s"))) ** 2  # Phi_i = pi/4
        R_th_pp = (np.abs(self.ReflectionCoeff(pi / 4, "p"))) ** 2

        # Incidence angle Phi_i = 0, 's' polarization

        # Incidence angle Phi_i = pi/4, 's' and 'p' polarizations
        data2 = self.s.evaluate(self.lbda_list, np.rad2deg(pi / 4))

        R_ss = data2.R_ss
        R_pp = data2.R_pp

        np.testing.assert_array_almost_equal(R_ss, R_th_ss, decimal=1)
        np.testing.assert_array_almost_equal(R_pp, R_th_pp, decimal=1)

    def test_normal_incidence_4x4_eig(self):
        data = self.s.evaluate(
            self.lbda_list, 0, solver=elli.Solver4x4, propagator=elli.PropagatorEig()
        )
        R_ss_0 = data.R_ss
        R_th_ss_0 = (np.abs(self.ReflectionCoeff(0, "s"))) ** 2  # Phi_i = 0
        np.testing.assert_allclose(R_ss_0, R_th_ss_0, 1e-10, 0)

    def test_angle_4x4_eig(self):
        R_th_ss = (np.abs(self.ReflectionCoeff(pi / 4, "s"))) ** 2  # Phi_i = pi/4
        R_th_pp = (np.abs(self.ReflectionCoeff(pi / 4, "p"))) ** 2

        # Incidence angle Phi_i = 0, 's' polarization

        # Incidence angle Phi_i = pi/4, 's' and 'p' polarizations
        data2 = self.s.evaluate(
            self.lbda_list,
            np.rad2deg(pi / 4),
            solver=elli.Solver4x4,
            propagator=elli.PropagatorEig(),
        )

        R_ss = data2.R_ss
        R_pp = data2.R_pp

        np.testing.assert_array_almost_equal(R_ss, R_th_ss, decimal=1)
        np.testing.assert_array_almost_equal(R_pp, R_th_pp, decimal=1)

    def test_angle_2x2(self):
        R_th_ss = (np.abs(self.ReflectionCoeff(pi / 4, "s"))) ** 2  # Phi_i = pi/4
        R_th_pp = (np.abs(self.ReflectionCoeff(pi / 4, "p"))) ** 2

        # Incidence angle Phi_i = 0, 's' polarization

        # Incidence angle Phi_i = pi/4, 's' and 'p' polarizations
        data2 = self.s.evaluate(
            self.lbda_list, np.rad2deg(pi / 4), solver=elli.Solver2x2
        )

        R_ss = data2.R_ss
        R_pp = data2.R_pp

        np.testing.assert_array_almost_equal(R_ss, R_th_ss, decimal=1)
        np.testing.assert_array_almost_equal(R_pp, R_th_pp, decimal=1)



# Encoding: utf-8
from abc import ABC, abstractmethod
from typing import Literal

import numpy as np
import numpy.typing as npt
import scipy.constants as sc

try:
    import torch
except ImportError:
    TORCH_AVAILABLE = False
else:
    TORCH_AVAILABLE = True

from numpy.lib.scimath import sqrt
from scipy.linalg import expm as scipy_expm

from .materials import IsotropicMaterial
from .result import Result
from .solver import Solver


class Propagator(ABC):
    """Propagator abstract base class."""

    @abstractmethod
    def calculate_propagation(
        self, delta: npt.NDArray, thickness: float, lbda: npt.ArrayLike
    ) -> npt.NDArray:
        """Calculates propagation for a given Delta matrix and layer thickness.

        Args:
            delta (npt.NDArray): Delta Matrix
            thickness (float): Thickness of layer (nm)
            lbda (npt.ArrayLike): Wavelengths to evaluate (nm)

        Returns:
            npt.NDArray: Propagator for the given layer
        """


class PropagatorLinear(Propagator):
    """Propagator class using a simple linear approximation of the matrix exponential."""

    def calculate_propagation(
        self, delta: npt.NDArray, thickness: float, lbda: npt.ArrayLike
    ) -> npt.NDArray:
        """Calculates propagation for a given Delta matrix and layer thickness with a linear approximation of the matrix exponential.

        Args:
            delta (npt.NDArray): Delta Matrix
            thickness (float): Thickness of layer (nm)
            lbda (npt.ArrayLike): Wavelengths to evaluate (nm)

        Returns:
            npt.NDArray: Propagator for the given layer
        """
        p_hs_lin = np.identity(4) + 1j * thickness * np.einsum(
            "nij,n->nij", delta, 2 * sc.pi / lbda
        )
        return p_hs_lin


class PropagatorExpm(Propagator):
    """Propagator class using the Padé approximation of the matrix exponential."""

    def __init__(self, backend: Literal["torch", "scipy", "automatic"] = "automatic"):
        """The Propagator can use two different backends: SciPy and PyTorch.
        The default installation only provides SciPy.
        PyTorch is faster and will be used automatically if available.
        If you want to install PyTorch please follow the instructions at https://pytorch.org/get-started/locally/.

        Args:
            backend (Literal["torch", "scipy", "automatic"], optional): Setting to change the linear algebra provider. Defaults to "automatic".
        """
        backends = {
            "torch": lambda mats: torch.linalg.matrix_exp(
                torch.from_numpy(mats)
            ).numpy(),
            "scipy": lambda mats: scipy_expm(mats),
        }

        if backend == "automatic" and TORCH_AVAILABLE:
            backend = "torch"
        elif backend == "automatic" and not TORCH_AVAILABLE:
            backend = "scipy"
        elif backend == "torch" and not TORCH_AVAILABLE:
            raise ImportError(
                "PyTorch is not installed. If you want to use the PyTorch backend, \
            please follow the install instructions on https://pytorch.org/get-started/locally/"
            )
        elif backend not in backends:
            raise ValueError(
                "Backend should be one of 'torch', 'scipy' or 'automatic'."
            )

        self.expm = backends[backend]

    def calculate_propagation(
        self, delta: npt.NDArray, thickness: float, lbda: npt.ArrayLike
    ) -> npt.NDArray:
        """Calculates propagation for a given Delta matrix and layer thickness with the Padé approximation of the matrix exponential.

        Args:
            delta (npt.NDArray): Delta Matrix
            thickness (float): Thickness of layer (nm)
            lbda (npt.ArrayLike): Wavelengths to evaluate (nm)

        Returns:
            npt.NDArray: Propagator for the given layer
        """
        mats = 1j * thickness * np.einsum("nij,n->nij", delta, 2 * sc.pi / lbda)

        propagator = self.expm(mats)

        return propagator


class PropagatorEig(Propagator):
    """Propagator class using the eigenvalue decomposition method."""

    def calculate_propagation(
        self, delta: npt.NDArray, thickness: float, lbda: npt.ArrayLike
    ) -> npt.NDArray:
        """Calculates propagation for a given Delta matrix and layer thickness with eigenvalue decomposition.

        Args:
            delta (npt.NDArray): Delta Matrix
            thickness (float): Thickness of layer (nm)
            lbda (npt.ArrayLike): Wavelengths to evaluate (nm)

        Returns:
            npt.NDArray: Propagator for the given layer
        """
        q, w = np.linalg.eig(delta)

        # Sort according to z propagation direction, by Re(q) first, then Im(q)
        i = np.lexsort((-np.real(q), -np.imag(q)))

        q = np.take_along_axis(q, i, axis=-1)
        w = np.take_along_axis(w, i[:, np.newaxis, :], axis=-1)

        w_i = np.linalg.inv(w)

        q = np.exp(q * 2j * thickness * sc.pi / lbda[:, None])

        p = np.zeros((lbda.shape[0], 4, 4), dtype=np.complex128)
        for i in range(4):
            p[:, i, i] = q[:, i]

        return w @ p @ w_i


class Solver4x4(Solver):
    """Solver class to evaluate Experiment objects. Based on Berreman's 4x4 method."""

    @staticmethod
    def build_delta_matrix(k_x: npt.ArrayLike, eps: npt.NDArray) -> npt.NDArray:
        """Calculates Delta matrix for given permittivity and reduced wave number.

        Args:
            k_x (npt.ArrayLike): reduce wave number, Kx = kx/k0
            eps (npt.NDArray): permittivity tensor

        Returns:
            npt.NDArray: Delta 4x4 matrix: infinitesimal propagation matrix
        """
        if np.shape(k_x) == ():
            length = 1
        else:
            length = np.shape(k_x)[0]

        zeros = np.tile(0, length)
        ones = np.tile(1, length)

        delta = np.array(
            [
                [
                    -k_x * eps[:, 2, 0] / eps[:, 2, 2],
                    -k_x * eps[:, 2, 1] / eps[:, 2, 2],
                    zeros,
                    ones - k_x**2 / eps[:, 2, 2],
                ],
                [zeros, zeros, -ones, zeros],
                [
                    eps[:, 1, 2] * eps[:, 2, 0] / eps[:, 2, 2] - eps[:, 1, 0],
                    k_x**2 - eps[:, 1, 1] + eps[:, 1, 2] * eps[:, 2, 1] / eps[:, 2, 2],
                    zeros,
                    k_x * eps[:, 1, 2] / eps[:, 2, 2],
                ],
                [
                    eps[:, 0, 0] - eps[:, 0, 2] * eps[:, 2, 0] / eps[:, 2, 2],
                    eps[:, 0, 1] - eps[:, 0, 2] * eps[:, 2, 1] / eps[:, 2, 2],
                    zeros,
                    -k_x * eps[:, 0, 2] / eps[:, 2, 2],
                ],
            ],
            dtype=np.complex128,
        )
        delta = np.moveaxis(delta, 2, 0)
        return delta

    @staticmethod
    def transition_matrix_halfspace(delta: npt.NDArray) -> npt.NDArray:
        """Returns transition exit matrix L for any half-space.

        Sort eigenvectors of the Delta matrix according to propagation
        direction first, then according to $y$ component.

        Returns eigenvectors ordered like (s+,s-,p+,p-)

        Args:
            delta (npt.NDArray): Delta 4x4 matrix: infinitesimal propagation matrix

        Returns:
            npt.NDArray: Translation matrix for semi-infinite half-spaces
        """
        q, p = np.linalg.eig(delta)

        # Sort according to z propagation direction, by Re(q) first, then Im(q)
        idx = np.lexsort((-np.real(q), -np.imag(q)))

        q = np.take_along_axis(q, idx, axis=-1)
        p = np.take_along_axis(p, idx[:, np.newaxis, :], axis=-1)
        # Result should be (+,+,-,-)

        # For each direction, sort according to Ey component, highest Ey first
        i1 = np.argsort(-np.abs(p[:, 1, :2]))
        i2 = 2 + np.argsort(-np.abs(p[:, 1, 2:]))
        i = np.hstack((i1, i2))
        # Result should be (s+,p+,s-,p-)

        # Reorder
        i[:, [1, 2]] = i[:, [2, 1]]

        q = np.take_along_axis(q, i, axis=-1)
        p = np.take_along_axis(p, i[:, np.newaxis, :], axis=-1)
        # Result should be(s+,s-,p+,p-)

        # Adjust Ey in ℝ⁺ for 's', and Ex in ℝ⁺ for 'p'
        e = np.hstack((p[:, 1, :2], p[:, 0, 2:]))

        ne = np.abs(e)
        c = np.ones_like(e)
        i = ne != 0.0
        c[i] = e[i] / ne[i]

        p = p * c[:, np.newaxis, :]

        # Normalize so that Ey = c1 + c2, analog to Ey = Eis + Ers
        # For an isotropic half-space, this should return the same matrix
        # as IsotropicHalfSpace
        c = p[:, 1, 0] + p[:, 1, 1]
        np.where(np.abs(c) == 0, 1, c)

        p = 2 * p / c[:, np.newaxis, np.newaxis]

        return p

    @staticmethod
    def transition_matrix_iso_halfspace(
        k_x: npt.ArrayLike, epsilon: npt.ArrayLike, inv: bool = False
    ) -> npt.NDArray:
        """Returns transition incident or exit matrix L for isotropic half-spaces.

        Args:
            k_x (npt.ArrayLike): Reduced wavenumber, Kx = kx/k0
            epsilon (npt.ArrayLike): dielectric tensor
            inv (bool, optional): If True, returns inverse transition matrix L^-1, used for the incident Matrix Li. Defaults to False.

        Returns:
            npt.NDArray: transition matrix L
        """
        n_x = sqrt(epsilon[:, 0, 0])
        sin_phi = k_x / n_x
        cos_phi = sqrt(1 - sin_phi**2)

        length = np.shape(k_x)[0]
        zeros = np.tile(0, length)
        ones = np.tile(1, length)

        if inv:
            sp_to_xy = np.array(
                [
                    [zeros, ones, -ones / cos_phi / n_x, zeros],
                    [zeros, ones, ones / cos_phi / n_x, zeros],
                    [ones / cos_phi, zeros, zeros, ones / n_x],
                    [-ones / cos_phi, zeros, zeros, ones / n_x],
                ],
                dtype=np.complex128,
            )
            return np.moveaxis(0.5 * sp_to_xy, 2, 0)

        sp_to_xy = np.array(
            [
                [zeros, zeros, cos_phi, -cos_phi],
                [ones, ones, zeros, zeros],
                [-n_x * cos_phi, n_x * cos_phi, zeros, zeros],
                [zeros, zeros, n_x, n_x],
            ],
            dtype=np.complex128,
        )
        return np.moveaxis(sp_to_xy, 2, 0)

    @staticmethod
    def get_k_z(
        material: "Material", lbda: npt.ArrayLike, k_x: npt.ArrayLike
    ) -> npt.NDArray:
        """Calculates Kz in a material

        Args:
            material (Material): Material of the half-space
            lbda (npt.ArrayLike): Wavelengths to evaluate (nm)
            k_x (npt.ArrayLike): Reduced wavenumber, Kx = kx/k0

        Returns:
            npt.NDArray: value of Kz in the material
        """
        nx = material.get_refractive_index(lbda)[:, 0, 0]
        k_z2 = nx**2 - k_x**2
        return sqrt(k_z2)

    def __init__(
        self, experiment: "Experiment", propagator: Propagator = PropagatorExpm()
    ) -> None:
        super().__init__(experiment)
        self.propagator = propagator

    def calculate(self) -> Result:
        """Calculates transition matrices for every element in the structure and resulting Jones matrices.

        Returns:
            Result: Result object with calculation results
        """
        # Kx = kx/k0 = n sin(Φ) : Reduced wavenumber.
        nx = self.structure.front_material.get_refractive_index(self.lbda)[:, 0, 0]
        k_x = nx * np.sin(np.deg2rad(self.theta_i))

        layers = reversed(self.permittivity_profile[1:-1])

        if isinstance(self.structure.back_material, IsotropicMaterial):
            m_t = self.transition_matrix_iso_halfspace(
                k_x, self.permittivity_profile[-1][1]
            )
        else:
            m_t = self.transition_matrix_halfspace(
                self.build_delta_matrix(k_x, self.permittivity_profile[-1][1])
            )

        for thickness, epsilon in layers:
            m_p = self.propagator.calculate_propagation(
                self.build_delta_matrix(k_x, epsilon), -thickness, self.lbda
            )
            m_t = m_p @ m_t

        m_lf = self.transition_matrix_iso_halfspace(
            k_x, self.permittivity_profile[0][1], inv=True
        )
        m_t = m_lf @ m_t

        # Extraction of t_it out of m_t. "2::-2" means integers {2,0}.
        t_it = m_t[:, 2::-2, 2::-2]
        # Calculate the inverse and make sure it is a matrix.
        t_ti = np.linalg.inv(t_it)

        # Extraction of t_rt out of m_t. "3::-2" means integers {3,1}.
        t_rt = m_t[:, 3::-2, 2::-2]

        # Then we have t_ri = t_rt * t_ti
        t_ri = t_rt @ t_ti

        jones_matrix_t = t_ti
        jones_matrix_r = t_ri

        # The power transmission coefficient is the ratio of the 'z' components
        # of the Poynting vector:       t = P_t_z / P_i_z
        # For isotropic media, we have: t = kb'/kf' |t_bf|^2
        # The correction coefficient is kb'/kf'
        # Note : For the moment it is only meaningful for isotropic half spaces.
        if isinstance(self.structure.back_material, IsotropicMaterial):
            k_z_f = self.get_k_z(self.structure.front_material, self.lbda, k_x)
            k_z_b = self.get_k_z(self.structure.back_material, self.lbda, k_x)
            power_correction = k_z_b.real / k_z_f.real
            return Result(
                self.experiment, jones_matrix_r, jones_matrix_t, power_correction
            )

        return Result(self.experiment, jones_matrix_r, jones_matrix_t)



"""Testing benchmark for each solver"""

import elli
import numpy as np
from elli.fitting import ParamsHist
from pytest import fixture


@fixture
def structure():
    """Build a structure"""
    params = ParamsHist()
    params.add("SiO2_n0", value=1.452, min=-100, max=100, vary=False)
    params.add("SiO2_n1", value=36.0, min=-40000, max=40000, vary=False)
    params.add("SiO2_n2", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_k0", value=0, min=-100, max=100, vary=False)
    params.add("SiO2_k1", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_k2", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_d", value=276.36, min=0, max=40000, vary=False)

    params.add("TiO2_n0", value=2.236, min=-100, max=100, vary=True)
    params.add("TiO2_n1", value=451, min=-40000, max=40000, vary=True)
    params.add("TiO2_n2", value=251, min=-40000, max=40000, vary=True)
    params.add("TiO2_k0", value=0, min=-100, max=100, vary=False)
    params.add("TiO2_k1", value=0, min=-40000, max=40000, vary=False)
    params.add("TiO2_k2", value=0, min=-40000, max=40000, vary=False)

    params.add("TiO2_d", value=20, min=0, max=40000, vary=True)

    SiO2 = elli.Cauchy(
        params["SiO2_n0"],
        params["SiO2_n1"],
        params["SiO2_n2"],
        params["SiO2_k0"],
        params["SiO2_k1"],
        params["SiO2_k2"],
    ).get_mat()
    TiO2 = elli.Cauchy(
        params["TiO2_n0"],
        params["TiO2_n1"],
        params["TiO2_n2"],
        params["TiO2_k0"],
        params["TiO2_k1"],
        params["TiO2_k2"],
    ).get_mat()

    Layer = [
        elli.Layer(TiO2, params["TiO2_d"]),
        elli.Layer(SiO2, params["SiO2_d"]),
        elli.Layer(TiO2, params["TiO2_d"]),
        elli.Layer(SiO2, params["SiO2_d"]),
        elli.Layer(TiO2, params["TiO2_d"]),
        elli.Layer(SiO2, params["SiO2_d"]),
        elli.Layer(TiO2, params["TiO2_d"]),
        elli.Layer(SiO2, params["SiO2_d"]),
    ]

    return elli.Structure(elli.AIR, Layer, elli.AIR)


lbda = np.linspace(400, 800, 500)
PHI = 70


def test_solver4x4_eig(benchmark, structure):
    """Benchmarks eignvalue propagator with solver4x4"""
    benchmark.pedantic(
        structure.evaluate,
        args=(lbda, PHI),
        kwargs={"solver": elli.Solver4x4, "propagator": elli.PropagatorEig()},
        iterations=1,
        rounds=10,
    )


def test_solver4x4_expm(benchmark, structure):
    """Benchmarks expm-scipy propagator with solver4x4"""
    benchmark.pedantic(
        structure.evaluate,
        args=(lbda, PHI),
        kwargs={
            "solver": elli.Solver4x4,
            "propagator": elli.PropagatorExpm(backend="scipy"),
        },
        iterations=1,
        rounds=10,
    )


def test_solver4x4_expm_pytorch(benchmark, structure):
    """Benchmarks expm-torch propagator with solver4x4"""
    benchmark.pedantic(
        structure.evaluate,
        args=(lbda, PHI),
        kwargs={
            "solver": elli.Solver4x4,
            "propagator": elli.PropagatorExpm(backend="torch"),
        },
        iterations=1,
        rounds=10,
    )


def test_solver4x4_linear(benchmark, structure):
    """Benchmarks linear propagator with solver4x4"""
    benchmark.pedantic(
        structure.evaluate,
        args=(lbda, PHI),
        kwargs={"solver": elli.Solver4x4, "propagator": elli.PropagatorLinear()},
        iterations=1,
        rounds=10,
    )


def test_solver2x2(benchmark, structure):
    """Benchmarks solver2x2"""
    benchmark.pedantic(
        structure.evaluate,
        args=(lbda, PHI),
        kwargs={"solver": elli.Solver2x2},
        iterations=1,
        rounds=10,
    )



"""Tests for a TiO2/SiO2/Si reference layer"""

from __future__ import unicode_literals

import os
from shutil import copytree, rmtree

import elli
import numpy as np
from elli.fitting import ParamsHist
from pytest import fixture


@fixture
def datadir(tmpdir, request):
    """
    Fixture responsible for searching a folder with the same name of test
    module and, if available, moving all contents to a temporary directory so
    tests can use them freely.
    """
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)

    if os.path.isdir(test_dir):
        rmtree(tmpdir)
        copytree(test_dir, str(tmpdir))

    return tmpdir


@fixture
def meas_data(datadir):
    """Fixture for getting the reference measurement data from the file."""
    return (
        elli.read_spectraray_rho(datadir.join("TiO2_400cycles.txt"))
        .loc[70.06]
        .loc[400:800]
    )


@fixture
def si_dispersion(datadir):
    """Fixture to load the silicon dispersion from file."""
    sr = elli.TableSpectraRay(datadir)
    silicon = elli.IsotropicMaterial(sr.load_dispersion_table(f"{os.sep}Si_Aspnes.mat"))

    return silicon


class TestTiO2:
    """Test cases for TiO2/SiO2/Si reference layers"""

    params = ParamsHist()
    params.add("SiO2_n0", value=1.452, min=-100, max=100, vary=False)
    params.add("SiO2_n1", value=36.0, min=-40000, max=40000, vary=False)
    params.add("SiO2_n2", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_k0", value=0, min=-100, max=100, vary=False)
    params.add("SiO2_k1", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_k2", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_d", value=276.36, min=0, max=40000, vary=False)

    params.add("TiO2_n0", value=2.23183200, min=-100, max=100, vary=True)
    params.add("TiO2_n1", value=449.066847, min=-40000, max=40000, vary=True)
    params.add("TiO2_n2", value=199.774450, min=-40000, max=40000, vary=True)
    params.add("TiO2_k0", value=0, min=-100, max=100, vary=False)
    params.add("TiO2_k1", value=0, min=-40000, max=40000, vary=False)
    params.add("TiO2_k2", value=0, min=-40000, max=40000, vary=False)

    params.add("TiO2_d", value=24.8772291, min=0, max=40000, vary=True)

    SiO2 = elli.IsotropicMaterial(
        elli.Cauchy(
            n0=params["SiO2_n0"],
            n1=params["SiO2_n1"],
            n2=params["SiO2_n2"],
            k0=params["SiO2_k0"],
            k1=params["SiO2_k1"],
            k2=params["SiO2_k2"],
        )
    )
    TiO2 = elli.IsotropicMaterial(
        elli.Cauchy(
            n0=params["TiO2_n0"],
            n1=params["TiO2_n1"],
            n2=params["TiO2_n2"],
            k0=params["TiO2_k0"],
            k1=params["TiO2_k1"],
            k2=params["TiO2_k2"],
        )
    )

    Layer = [elli.Layer(TiO2, params["TiO2_d"]), elli.Layer(SiO2, params["SiO2_d"])]

    @staticmethod
    def chisqr(meas_data, sim_data):
        """Calculates the chi square value between two arrays.

        Args:
            meas_data (pd.DataFrame): Measurement data array
            sim_data (pd.DataFrame): Simulation data array

        Returns:
            pd.DataFrame:
                The chi square difference between the measurement and simulation array data.
        """
        chisqr_real = ((np.real(meas_data) - np.real(sim_data)) ** 2).sum()
        chisqr_imag = ((np.imag(meas_data) - np.imag(sim_data)) ** 2).sum()

        return chisqr_real + chisqr_imag

    def test_solver2x2(self, si_dispersion, meas_data):
        """The solver2x2 is within chi square accuracy"""
        sim_data = (
            elli.Structure(elli.AIR, self.Layer, si_dispersion)
            .evaluate(meas_data.index, 70, solver=elli.Solver2x2)
            .rho
        )

        assert TestTiO2.chisqr(meas_data, sim_data) < 0.0456

    def test_solver4x4_expm(self, si_dispersion, meas_data):
        """The solver4x4 with scipy propagator is within chi square accuracy"""
        sim_data = (
            elli.Structure(elli.AIR, self.Layer, si_dispersion)
            .evaluate(
                meas_data.index,
                70,
                solver=elli.Solver4x4,
                propagator=elli.PropagatorExpm(backend="scipy"),
            )
            .rho
        )

        assert TestTiO2.chisqr(meas_data, sim_data) < 0.0456

    def test_solver4x4_expm_torch(self, si_dispersion, meas_data):
        """The solver4x4 with pytorch propagator is within chi square accuracy"""
        sim_data = (
            elli.Structure(elli.AIR, self.Layer, si_dispersion)
            .evaluate(
                meas_data.index,
                70,
                solver=elli.Solver4x4,
                propagator=elli.PropagatorExpm(backend="torch"),
            )
            .rho
        )

        assert TestTiO2.chisqr(meas_data, sim_data) < 0.0456

    def test_solver4x4_eig(self, si_dispersion, meas_data):
        """The solver4x4 with eig propagator is within chi square accuracy"""
        sim_data = (
            elli.Structure(elli.AIR, self.Layer, si_dispersion)
            .evaluate(
                meas_data.index,
                70,
                solver=elli.Solver4x4,
                propagator=elli.PropagatorEig(),
            )
            .rho
        )

        assert TestTiO2.chisqr(meas_data, sim_data) < 0.0456



"""Testing cases for the dispersion law factory class."""

import elli
import pytest


def test_correct_object_generation():
    """Factory class creates a correct object"""
    factory_disp_less = elli.DispersionFactory.get_dispersion(
        "ConstantRefractiveIndex", n=1.5
    )

    assert isinstance(factory_disp_less, elli.ConstantRefractiveIndex)
    assert factory_disp_less.single_params.get("n") == 1.5


def test_error_on_not_existing_class():
    """Raises an error if dispersion does not exist."""
    with pytest.raises(ValueError):
        elli.DispersionFactory.get_dispersion("DispersionNotExisting", n=1000)


def test_error_on_bad_class():
    """Raises an error if a bad class is requested."""
    for bad_class in ["DispersionFactory", "Dispersion", "DispersionSum"]:
        with pytest.raises(ValueError):
            elli.DispersionFactory.get_dispersion(bad_class)



"""Tests for the result class"""

import elli
import numpy as np
from pytest import fixture, raises


@fixture
def result():
    """Generate a pyElli result object"""
    Si = elli.db.RII().get_mat("Si", "Aspnes")
    SiO2 = elli.Cauchy(1.452, 36.0).get_mat()
    structure = elli.Structure(elli.AIR, [elli.Layer(SiO2, 500)], Si)

    return structure.evaluate(np.linspace(250, 800), 70, solver=elli.Solver2x2)


def test_delta_range_conversion(result):
    """Checks correct delta range conversion"""
    assert (
        all(result.delta > -180) and all(result.delta < 180) and any(result.delta < 0)
    )

    result.as_delta_range(0, 360)
    assert all(result.delta > 0) and all(result.delta < 360)

    result.as_delta_range(0, 180)
    assert all(result.delta > 0) and all(result.delta < 180)

    result.as_delta_range(-180, 180)
    assert (
        all(result.delta > -180) and all(result.delta < 180) and any(result.delta < 0)
    )


def test_delta_range_error_on_invalid_range(result):
    """Checks raising of errors on invalid range or type."""
    with raises(TypeError):
        result.as_delta_range("hallo", "welt")

    with raises(ValueError):
        result.as_delta_range(20, 180)


def test_resultlist_shape(result):
    result_list = elli.ResultList([result, result, result])

    assert len(result_list) == 3
    assert np.shape(result_list.delta) == (3, 50)

    assert np.shape(result_list.mean.delta) == (50,)
    assert np.allclose(result_list.mean.delta, result.delta)



r"""Calculate Kramers-Kronig relations according to Maclaurin's formula [1]_.
Here, the differential formulation is used,
which means that the transformation always diverges to zero at infinity,
leaving a free infinity offset.
This offset is typically referred to as :math:`\epsilon(\infty)` in spectroscopy.

Since the Kramers-Kronig relation integrates over the whole spectrum for each point,
it is very sensitive to sampling changes and non-zero values.
To obtain best values the y-axis
should be zero anywhere outside the integration range.
However, since this is not possible for every dispersion
relation the calculation should be done on a much wider
range than the actual interval used.
Additionally, the discretisation steps of the x-axis should be kept constant.

For the transformation of the real to imaginary part you need to be especially cautios.
Typically, in spectroscopy the real part of the dielectric
function is non-zero throughout the whole spectrum
and especially :math:`\epsilon(\infty) \ne 0`.
This makes the Kramers-Kronig transformation virtually impossible in these cases
as it suffers from major uncertanties.
Hence, this transformation is included in pyElli only for completeness and for the
special cases it may be applicable.

.. rubric:: References

.. [1] Ohta and Ishida, Appl. Spectroscopy 42, 952 (1988), https://doi.org/10.1366/0003702884430380
"""

# pylint: disable=invalid-name
from typing import Callable

import numpy as np


def _integrate_im(im: np.ndarray, x: np.ndarray, x_i: np.ndarray) -> np.ndarray:
    """Calculate the discrete imaginary sum (integral) for the kkr.

    Args:
        im (numpy.ndarray): The imaginary values from which to calculate. (shape (1, n))
        x (numpy.ndarray): The x-axis on which to calculate. (shape (1, n))
        x_i (numpy.ndarray): The current points around which to integrate. (shape (m, 1))

    Returns:
        numpy.ndarray: The integral sum. (shape (m,))
    """

    return np.sum(x * im / (x * x - x_i * x_i), axis=1)


def _integrate_im_reciprocal(
    im: np.ndarray, x: np.ndarray, x_i: np.ndarray
) -> np.ndarray:
    """Calculate the discrete imaginary sum (integral) for the kkr.
    This formulation uses an 1/x axis to transform a wavelength axis.

    Args:
        im (numpy.ndarray): The imaginary values from which to calculate. (shape (1, n))
        x (numpy.ndarray): The reciprocal x-axis on which to calculate. (shape (1, n))
        x_i (numpy.ndarray): The current point around which to integrate. (shape (m, 1))

    Returns:
        numpy.ndarray: The integral sum. (shape (m,))
    """

    return np.sum(im / (x * (1.0 - x * x / (x_i * x_i))), axis=1)


def _integrate_re(re: np.ndarray, x: np.ndarray, x_i: np.ndarray) -> np.ndarray:
    """Calculate the discrete real sum (integral) for the kkr.

    Args:
        re (numpy.ndarray): The real values from which to calculate. (shape (1, n))
        x (numpy.ndarray): The x-axis on which to calculate. (shape (1, n))
        x_i (numpy.ndarray): The current point around which to integrate. (shape (m, 1))

    Returns:
        numpy.ndarray: The real sum. (shape (m,))
    """
    return np.sum(x_i * re / (x * x - x_i * x_i), axis=1)


def _integrate_re_reciprocal(
    re: np.ndarray, x: np.ndarray, x_i: np.ndarray
) -> np.ndarray:
    """Calculate the discrete real sum (integral) for the kkr.
    This formulation uses an 1/x axis to transform a wavelength axis.

    Args:
        re (numpy.ndarray): The real values from which to calculate. (shape (1, n))
        x (numpy.ndarray): The reciprocal x-axis on which to calculate. (shape (1, n))
        x_i (float): The current point around which to integrate. (shape (m, 1))

    Returns:
        numpy.ndarray: The real sum. (shape (m,))
    """

    return np.sum(re / (x_i - x * x / x_i), axis=1)


def _calc_kkr(
    t: np.ndarray,
    x: np.ndarray,
    trafo: Callable[[np.ndarray, np.ndarray, np.ndarray], np.ndarray],
) -> np.ndarray:
    """Calculates the Kramers-Kronig relation
    according to Maclaurin's formula.

    Args:
        t (numpy.ndarray): The y-axis on which to transform.
        x (numpy.ndarray): The x-axis on which to transform.
        trafo (Callable[[numpy.ndarray, numpy.ndarray, numpy.ndarray], numpy.ndarray]):
            The transformation function.

    Raises:
        ValueError: y and x axis must have the same length.

    Returns:
        np.ndarray: The kkr transformed y-axis
    """

    if len(t) != len(x):
        raise ValueError(
            "y- and x-axes arrays must have the same length, "
            f"but have lengths {len(t)} and {len(x)}."
        )

    integral = np.empty(len(t))
    interval = np.diff(x, prepend=x[1] - x[0])
    odd_slice = slice(1, None, 2)
    even_slice = slice(0, None, 2)

    integral[even_slice] = trafo(
        t[np.newaxis, odd_slice],
        x[np.newaxis, odd_slice],
        x[even_slice, np.newaxis],
    )
    integral[odd_slice] = trafo(
        t[np.newaxis, even_slice],
        x[np.newaxis, even_slice],
        x[odd_slice, np.newaxis],
    )

    return 4 / np.pi * interval * integral


def re2im(re: np.ndarray, x: np.ndarray) -> np.ndarray:
    r"""Calculates the differential Kramers-Kronig relation from the
    real to imaginary part
    according to Maclaurin's formula.

    The underlying formula reads:

    .. math::
        \Delta \Im(x_i) = \Im(x_i) - \Im(\infty) =
        \frac{2}{\pi} \int_0^\infty \frac{x_i \Re(x)}{x^2 - x_i^2} dx

    Args:
        re (numpy.ndarray): The real values to transform.
        x (numpy.ndarray): The axis on which to transform.

    Returns:
        numpy.ndarray: The transformed imaginary part.
    """

    return _calc_kkr(re, x, _integrate_re)


def im2re(im: np.ndarray, x: np.ndarray) -> np.ndarray:
    r"""Calculates the differential Kramers-Kronig relation from the
    imaginary to real part
    according to Maclaurin's formula.

    The underlying formula reads:

    .. math::
        \Delta \Re(x_i) = \Re(x_i) - \Re(\infty) =
        \frac{2}{\pi} \int_0^\infty \frac{x \Im(x)}{x^2 - x_i^2} dx

    Args:
        im (numpy.ndarray): The imaginary values to transform.
        x (numpy.ndarray): The axis on which to transform.

    Returns:
        numpy.ndarray: The transformed real part.
    """

    return _calc_kkr(im, x, _integrate_im)


def re2im_reciprocal(re: np.ndarray, x: np.ndarray) -> np.ndarray:
    r"""Calculates the differential Kramers-Kronig relation from the
    real to imaginary part
    according to Maclaurin's formula.
    This function assumes a reciprocal x-axis, e.g. wavelength in spectroscopy.

    The underlying formula reads:

    .. math::
        \Delta \Im(x_i) = \Im(x_i) - \Im(\infty) =
        \frac{2}{\pi} \int_0^\infty \frac{\Re(x)}{x_i - \frac{x^2}{x_i}} dx

    Args:
        re (numpy.ndarray): The real values to transform.
        x (numpy.ndarray): The reciprocal axis on which to transform.

    Returns:
        numpy.ndarray: The transformed imaginary part.
    """

    return _calc_kkr(re, x, _integrate_re_reciprocal)


def im2re_reciprocal(im: np.ndarray, x: np.ndarray) -> np.ndarray:
    r"""Calculates the differential Kramers-Kronig relation from the
    imaginary to real part
    according to Maclaurin's formula.
    This function assumes a reciprocal x-axis, e.g. wavelength in spectroscopy.

     The underlying formula reads:

    .. math::
        \Delta \Re(x_i) = \Re(x_i) - \Re(\infty) =
        \frac{2}{\pi} \int_0^\infty \frac{x \Im(x)}{1 - \frac{x^2}{x_i^2}} dx

    Args:
        im (numpy.ndarray): The imaginary values to transform.
        x (numpy.ndarray): The reciprocal axis on which to transform.

    Returns:
        numpy.ndarray: The transformed real part.
    """

    return _calc_kkr(im, x, _integrate_im_reciprocal)



"""Tests for the different mixture models"""

import pytest
import elli
import numpy as np


material_list = [
    ("Si", "Aspnes"),
    ("GaAs", "Aspnes"),
    ("SiO2", "Malitson"),
    ("CaF2", "Li"),
    ("Rh", "Weaver"),
    ("MoS2", "Song-1L"),
]


@pytest.fixture
def RII():
    RII = elli.db.RII()
    return RII


@pytest.mark.parametrize(
    "host_book, host_page",
    material_list,
)
@pytest.mark.parametrize(
    "guest_book, guest_page",
    material_list,
)
@pytest.mark.parametrize(
    "EMA",
    [
        elli.LooyengaEMA,
        elli.MaxwellGarnettEMA,
    ],
)
def test_mixture_models(host_book, host_page, guest_book, guest_page, EMA, RII):
    host_mat = RII.get_mat(host_book, host_page)
    guest_mat = RII.get_mat(guest_book, guest_page)

    lbda = np.arange(250, 800, 25)

    np.testing.assert_allclose(
        elli.BruggemanEMA(host_mat, guest_mat, 0.1).get_tensor(lbda),
        EMA(host_mat, guest_mat, 0.1).get_tensor(lbda),
        rtol=0.5,
        atol=0.5 + 0.5j,
    )



# Encoding: utf-8
"""Cody-Lorentz dispersion law. Model by Ferlauto et al."""

from typing import Dict
import numpy as np
import numpy.typing as npt
from scipy.interpolate import interp1d

from ..utils import conversion_wavelength_energy
from .base_dispersion import Dispersion
from ..kkr import im2re_reciprocal


class CodyLorentz(Dispersion):
    """Cody-Lorentz dispersion law. Model by Ferlauto et al.

    Single parameters:
        :Eg: Bandgap energy (eV). Defaults to 1.6.
        :A: Amplitude (eV). Defaults to 100.
        :Et: Energy at which the Urbach tail starts (eV). Defaults to 1.8.
        :gamma: Broadening (eV). Defaults to 2.4.
        :Ep:
            Distance from bandgap for transition from Cody type absorption
            to Lorentz type absorption (eV). Defaults to 0.8.
        :E0: Lorentz resonance energy (eV). Defaults to 3.6.
        :Eu: Exponential decay of the Urbach tail (eV). Defaults to 0.05.

    Repeated parameters:
        --

    Output:
        The Cody-Lorentz dispersion. Please refer to the references for a full formula.

    References:
        * Ferlauto et al., J. Appl. Phys. 92, 2424 (2002)
    """

    single_params_template = {
        "Eg": 1.6,
        "A": 100,
        "Et": 1.8,
        "gamma": 2.4,
        "Ep": 0.8,
        "E0": 3.6,
        "Eu": 0.05,
    }
    rep_params_template: Dict[str, float] = {}

    @staticmethod
    def eps2(E, Eg, A, Et, gamma, Ep, E0, Eu):
        """The imaginary part of the cody lorentz dispersion"""

        # pylint: disable=invalid-name
        def G(E):
            return (E - Eg) ** 2 / ((E - Eg) ** 2 + Ep**2)

        def L(E):
            return A * E0 * gamma * E / ((E**2 - E0**2) ** 2 + gamma**2 * E**2)

        E1 = Et * G(Et) * L(Et)

        # fmt: off
        return (
            E1 / E * np.exp((E - Et) / Eu) * np.heaviside(Et - E, 1)
            + G(E) * L(E) * np.heaviside(E - Et, 0)
        )
        # fmt: on

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        energy = conversion_wavelength_energy(lbda)

        lbda_broad = np.linspace(50, 10000, 1000)
        energy_padded = conversion_wavelength_energy(lbda_broad)
        eps1 = im2re_reciprocal(
            CodyLorentz.eps2(energy_padded, **self.single_params), lbda_broad
        )
        eps1_interp = interp1d(lbda_broad, eps1)(lbda)

        return eps1_interp + 1j * CodyLorentz.eps2(energy, **self.single_params)



"""Tests for the formula dispersion"""

import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest

from benchmark_formula_dispersion import structure as formula_structure
from benchmark_propagators_TiO2 import structure
import elli
from elli.dispersions import Sellmeier, Formula
from elli.dispersions.cauchy import Cauchy
from elli.dispersions.formula import FormulaIndex


@pytest.mark.parametrize(
    "ref_model, formula_model, formula, single_params, rep_params, unit",
    [
        (
            Sellmeier,
            Formula,
            "eps = 1 + sum[A * (lbda * 1e-3)**2 / ((lbda * 1e-3)  ** 2 - B)]",
            {},
            {"A": [1, 1, 1], "B": [0.1, 0.1, 0.1]},
            "nm",
        ),
        (
            Sellmeier,
            Formula,
            "eps = 1 + sum[A * lbda**2 / (lbda** 2 - B)]",
            {},
            {"A": [1, 1, 1], "B": [0.1, 0.1, 0.1]},
            "micrometer",
        ),
        (
            Cauchy,
            FormulaIndex,
            (
                "n = n0 + 1e2 * n1 / lbda ** 2 + 1e7 * n2 / lbda ** 4 + "
                "1j * (k0 + 1e2 * k1 / lbda ** 2 + 1e7 * k2 / lbda ** 4)"
            ),
            {"n0": 1.5, "n1": 1, "n2": 1, "k0": 1, "k1": 1, "k2": 1},
            {},
            "nm",
        ),
    ],
)
def test_formula_reproduces_predefined(
    ref_model, formula_model, formula, single_params, rep_params, unit
):
    """The formula dispersion reproduces other dispersion models"""
    lbda = np.linspace(400, 1500, 500)

    disp = ref_model(**single_params)
    for params in zip(*rep_params.values()):
        disp.add(*params)

    formula_disp = formula_model(formula, "lbda", single_params, rep_params, unit)

    assert_array_almost_equal(
        disp.get_dielectric(lbda), formula_disp.get_dielectric(lbda)
    )


def test_formula_fails_on_wrong_repr():
    """The formula dispersion fails when it is tried to be
    initialized with the wrong represenation"""

    with pytest.raises(ValueError):
        Formula("n = 1", "", {}, {})

    with pytest.raises(ValueError):
        FormulaIndex("eps = 1", "", {}, {})


def test_formula_against_predefined_model(structure, formula_structure):
    """A formula dispersion evaluates to the same values as a predefined model"""
    lbda = np.linspace(400, 800, 500)
    PHI = 70

    predefined4x4 = structure.evaluate(
        lbda, PHI, solver=elli.Solver4x4, propagator=elli.PropagatorExpm()
    )
    formula4x4 = formula_structure.evaluate(
        lbda, PHI, solver=elli.Solver4x4, propagator=elli.PropagatorExpm()
    )

    assert_array_almost_equal(predefined4x4.rho, formula4x4.rho)

    predefined2x2 = structure.evaluate(lbda, PHI, solver=elli.Solver2x2)
    formula2x2 = formula_structure.evaluate(lbda, PHI, solver=elli.Solver2x2)

    assert_array_almost_equal(predefined2x2.rho, formula2x2.rho)



# Encoding: utf-8
"""Drude dispersion model with parameters in units of energy."""

import numpy.typing as npt

from ..utils import conversion_wavelength_energy
from .base_dispersion import Dispersion


class DrudeEnergy(Dispersion):
    r"""Drude dispersion model with parameters in units of energy.
    Drude models in the literature typically contain an additional epsilon infinity value.
    Use `EpsilonInf` to add this parameter or simply add a number, e.g. DrudeEnergy() + 2, where
    2 is the value of epsilon infinity.

    Single parameters:
        :A: Amplitude of Drude oscillator. Defaults to 0. Unit in eV\ :sup:`2`
        :gamma: Broadening of Drude oscillator. Defaults to 0. Unit in eV.

    Repeated parameters:
        --

    Output:
        .. math::
            \varepsilon(E)
            = \boldsymbol{A} / (E^2 - i \cdot \boldsymbol{gamma} \cdot E)
    """

    single_params_template = {"A": 0, "gamma": 0}
    rep_params_template = {}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        energy = conversion_wavelength_energy(lbda)
        return self.single_params.get("A") / (
            energy**2 - 1j * self.single_params.get("gamma") * energy
        )



#!/usr/bin/python3
# encoding: utf-8
# %% [markdown]
# # We reproduce results from section 6.4.1 in 'Spectroscopic Ellipsometry', by H. Fujiwara.
#
# Author: O. Castany, M. Müller
#
# Verification of the code against results presented in Fujiwara's book 'Spectroscopic Ellipsometry', section 6.4.1
# (p. 237). We reproduce figures 6.16 and 6.17.

# %%
import numpy as np
import elli
import matplotlib.pyplot as plt

# %% [markdown]
# ## Setting up the materials

# %%
# Front half-space (air)
air = elli.AIR

# Anisotropic substrate
n_o = 2.0  #  ordinary index of thin layer
n_e = 2.5  #  extraordinary index of thin layer

uniaxialMaterial = elli.UniaxialMaterial(
    elli.ConstantRefractiveIndex(n_o), elli.ConstantRefractiveIndex(n_e)
)

# %% [markdown]
# ## We reproduce figure 6.16 (p. 238)
# %%
# Orientations of the anisotropic substrate
Φ_E = 90  #  1st Euler angle
θ_E_list = [0, 45, 90]  #  2nd Eulet angle

# Incidence angles
Φ_i_list = np.linspace(0, 89, 300)  #  array of Φ_i values

data = elli.ResultList()

for θ_E in θ_E_list:
    R = elli.rotation_euler(Φ_E, θ_E, 0)
    uniaxialMaterial.set_rotation(R)
    s = elli.Structure(air, [], uniaxialMaterial)
    for Φ_i in Φ_i_list:
        data.append(s.evaluate(500, Φ_i))

Psi = data.psi_pp.reshape(3, 300).T

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(Φ_i_list, Psi)
ax.set_xlabel(r"$\theta_i$")
ax.set_ylabel(r"$\Psi_{pp}$")

plt.tight_layout()

# %% [markdown]
# ## We reproduce figure 6.17 (p. 239)

# %%
Φ_i = 70  #  70° incidence angle

# Orientations of the anisotropic substrate
Φ_E_list = np.linspace(0, 360, 36 * 2 + 1)  #  1st Euler angle
θ_E_list = [0, 45, 90]  #  2nd Euler angle

data2 = elli.ResultList()

for θ_E in θ_E_list:
    for Φ_E in Φ_E_list:
        R = elli.rotation_euler(Φ_E, θ_E, 0)
        uniaxialMaterial.set_rotation(R)
        s = elli.Structure(air, [], uniaxialMaterial)
        data2.append(s.evaluate(500, Φ_i))

Psi = data2.psi_pp.reshape(3, 73).T

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(Φ_E_list, Psi)
ax.set_xlim(0, 360)
ax.set_xlabel(r"$\phi_E$")
ax.set_ylabel(r"$\Psi_{pp}$")

plt.tight_layout()
plt.show()

# %%
